from typing import List, Dict, Set
from pathlib import PurePosixPath
from collections import deque
import re

from alipcs_py.alipcs import AliPCSApi, PcsFile
from alipcs_py.commands.list_files import list_files
from alipcs_py.commands.sifter import Sifter
from alipcs_py.commands.display import (
    display_invalid_shared_link_infos,
    display_shared_files,
    display_shared_link_infos,
    display_shared_links,
)
from alipcs_py.commands.download import (
    download,
    Downloader,
    DEFAULT_DOWNLOADER,
    DownloadParams,
    DEFAULT_DOWNLOADPARAMS,
)
from alipcs_py.commands.play import play, Player, DEFAULT_PLAYER

import requests  # type: ignore

from rich import print


def share_files(api: AliPCSApi, *remotepaths: str, password: str = "", period: int = 0):
    pcs_files = api.paths(*remotepaths)
    assert all(pcs_files)

    file_ids = [pf.file_id for pf in pcs_files if pf]

    shared_link = api.share(*file_ids, password=password, period=period)
    display_shared_links(shared_link)


def list_shared(api: AliPCSApi, show_all=True):
    pcs_shared_links = api.list_shared_all()
    if not pcs_shared_links:
        return

    display_shared_links(*pcs_shared_links)


def cancel_shared(api: AliPCSApi, *share_ids: str):
    api.cancel_shared(*share_ids)


def _redirect(url: str) -> str:
    if "alywp.net" not in url:
        return url
    resp = requests.get(url, allow_redirects=False)
    return resp.headers.get("Location") or ""


def _extract_share_id(share_url: str) -> str:
    m = re.search(r"/s/(\w+)", share_url)
    return m.group(1) if m else ""


def _extract_file_id(share_url: str) -> str:
    m = re.search(r"/folder/(\w+)", share_url)
    return m.group(1) if m else ""


def save_shared_by_url(api: AliPCSApi, remotedir: str, share_url: str, password: str = ""):
    share_url = _redirect(share_url)
    share_id = _extract_share_id(share_url)
    file_id = _extract_file_id(share_url)
    file_ids = [file_id] if file_id else []

    assert share_id

    save_shared_by_file_ids(api, remotedir, share_id, file_ids, password=password)


def save_shared_by_file_ids(
    api: AliPCSApi,
    remotedir: str,
    share_id: str,
    file_ids: List[str],
    password: str = "",
):
    assert share_id

    api.get_share_token(share_id, share_password=password)

    file_ids = file_ids or ["root"]

    sfs = api.meta(*file_ids, share_id=share_id)
    for sf in sfs:
        if not sf.path:
            sf.path = sf.name
    shared_files = deque(sfs)

    # Record the remotedir of each shared_file
    _remotedirs: Dict[str, str] = {}
    for sp in shared_files:
        _remotedirs[sp.file_id] = remotedir

    # Map the remotedir to its pcs_file
    dest_pcs_files: Dict[str, PcsFile] = {}

    while shared_files:
        shared_file = shared_files.popleft()
        rd = _remotedirs[shared_file.file_id]

        # Make sure remote dir exists
        if rd not in dest_pcs_files:
            dest_pcs_files[rd] = api.makedir_path(rd)
        dest_pcs_file = dest_pcs_files[rd]

        if not shared_file.is_root() and not remotepath_exists(api, shared_file.name, rd):
            api.transfer_shared_files(
                [shared_file.file_id],
                dest_pcs_file.file_id,
                share_id,
                auto_rename=False,
            )
            print(f"save: `{shared_file.path}` to `{rd}`")
        else:
            # Ignore existed file
            if shared_file.is_file:
                print(f"[yellow]WARNING[/]: `{shared_file.path}` has be in `{rd}`")
                continue
            else:  # shared_file.is_dir
                sub_files = list(api.list_path_iter(shared_file.path, file_id=shared_file.file_id, share_id=share_id))

                rd = (PurePosixPath(rd) / shared_file.name).as_posix()
                for sp in sub_files:
                    _remotedirs[sp.file_id] = rd
                shared_files.extendleft(sub_files[::-1])


def save_shared(
    api: AliPCSApi,
    remotedir: str,
    share_id: str = "",
    share_url: str = "",
    file_ids: List[str] = [],
    password: str = "",
):
    assert remotedir.startswith("/"), "`remotedir` must be an absolute path"

    assert int(bool(share_id)) ^ int(bool(share_url)), "`share_id` and `share_url` only can be given one"

    if share_url:
        save_shared_by_url(api, remotedir, share_url, password=password)
    else:
        save_shared_by_file_ids(api, remotedir, share_id, file_ids, password=password)


def list_shared_files(
    api: AliPCSApi,
    *remotepaths: str,
    share_id: str = "",
    share_url: str = "",
    password: str = "",
    file_ids: List[str] = [],
    desc: bool = False,
    name: bool = False,
    time: bool = False,
    size: bool = False,
    all: bool = True,
    limit: int = 200,
    recursive: bool = False,
    sifters: List[Sifter] = [],
    highlight: bool = False,
    show_size: bool = False,
    show_date: bool = False,
    show_file_id: bool = False,
    show_hash: bool = False,
    show_absolute_path: bool = False,
    csv: bool = False,
):
    assert int(bool(share_id)) ^ int(bool(share_url)), "`share_id` and `share_url` only can be given one"

    share_url = _redirect(share_url)

    if share_url:
        share_id = _extract_share_id(share_url)
        if not file_ids:
            file_id = _extract_file_id(share_url)
            file_ids = [file_id] if file_id else []

    if not remotepaths and not file_ids:
        return

    assert share_id

    api.get_share_token(share_id, share_password=password)

    list_files(
        api,
        *remotepaths,
        file_ids=file_ids,
        share_id=share_id,
        desc=desc,
        name=name,
        time=time,
        size=size,
        all=all,
        limit=limit,
        recursive=recursive,
        sifters=sifters,
        highlight=highlight,
        show_size=show_size,
        show_date=show_date,
        show_file_id=show_file_id,
        show_hash=show_hash,
        show_absolute_path=show_absolute_path,
        csv=csv,
    )


def remotepath_exists(api: AliPCSApi, name: str, rd: str, _cache: Dict[str, Set[str]] = {}) -> bool:
    names = _cache.get(rd)
    if not names:
        names = set([sp.name for sp in api.list_path_iter(rd)])
        _cache[rd] = names
    return name in names


def download_shared(
    api: AliPCSApi,
    remotepaths: List[str],
    file_ids: List[str],
    localdir: str,
    share_id: str = "",
    share_url: str = "",
    password: str = "",
    sifters: List[Sifter] = [],
    recursive: bool = False,
    from_index: int = 0,
    downloader: Downloader = DEFAULT_DOWNLOADER,
    downloadparams: DownloadParams = DEFAULT_DOWNLOADPARAMS,
    out_cmd: bool = False,
    encrypt_password: bytes = b"",
):
    assert int(bool(share_id)) ^ int(bool(share_url)), "`share_id` and `share_url` only can be given one"

    share_url = _redirect(share_url)

    if share_url:
        share_id = _extract_share_id(share_url)
        if not file_ids:
            file_id = _extract_file_id(share_url)
            file_ids = [file_id] if file_id else []

    if not remotepaths and not file_ids:
        return

    assert share_id

    api.get_share_token(share_id, share_password=password)

    download(
        api,
        remotepaths,
        file_ids=file_ids,
        localdir=localdir,
        share_id=share_id,
        sifters=sifters,
        recursive=recursive,
        from_index=from_index,
        downloader=downloader,
        downloadparams=downloadparams,
        out_cmd=out_cmd,
        encrypt_password=encrypt_password,
    )


def play_shared(
    api: AliPCSApi,
    remotepaths: List[str],
    file_ids: List[str],
    share_id: str,
    share_url: str = "",
    password: str = "",
    sifters: List[Sifter] = [],
    recursive: bool = False,
    from_index: int = 0,
    player: Player = DEFAULT_PLAYER,
    player_params: List[str] = [],
    quiet: bool = False,
    shuffle: bool = False,
    ignore_ext: bool = False,
    out_cmd: bool = False,
    local_server: str = "",
):
    assert int(bool(share_id)) ^ int(bool(share_url)), "`share_id` and `share_url` only can be given one"

    share_url = _redirect(share_url)

    if share_url:
        share_id = _extract_share_id(share_url)
        if not file_ids:
            file_id = _extract_file_id(share_url)
            file_ids = [file_id] if file_id else []

    if not remotepaths and not file_ids:
        return

    assert share_id

    api.get_share_token(share_id, share_password=password)

    play(
        api,
        remotepaths,
        file_ids=file_ids,
        share_id=share_id,
        sifters=sifters,
        recursive=recursive,
        from_index=from_index,
        player=player,
        player_params=player_params,
        quiet=quiet,
        shuffle=shuffle,
        ignore_ext=ignore_ext,
        out_cmd=out_cmd,
        local_server=local_server,
    )


def get_share_token(api: AliPCSApi, share_id: str, share_url: str = "", password: str = "") -> str:
    assert int(bool(share_id)) ^ int(bool(share_url)), "`share_id` and `share_url` only can be given one"

    share_url = _redirect(share_url)

    if share_url:
        share_id = _extract_share_id(share_url)

    assert share_id

    return api.get_share_token(share_id, share_password=password)



import os

from setuptools import setup
from Cython.Build import cythonize
from distutils.command.build_ext import build_ext


def build(setup_kwargs):
    extensions = ["alipcs_py/common/simple_cipher.pyx"]

    # gcc arguments hack: enable optimizations
    os.environ["CFLAGS"] = "-O3"

    ext_modules = cythonize(
        extensions,
        language_level=3,
        compiler_directives={"linetrace": True},
    )

    ext_modules[0].name = "alipcs_py.common.simple_cipher"

    # Build
    setup_kwargs.update(
        {
            "ext_modules": ext_modules,
            "cmdclass": {"build_ext": build_ext},
        }
    )


if __name__ == "__main__":
    setup_kwargs = {}
    build(setup_kwargs)
    setup(**setup_kwargs)



from typing import Optional

from alipcs_py.alipcs import AliPCSApi
from alipcs_py.commands.display import display_blocked_remotepath

import chardet  # type: ignore


def cat(
    api: AliPCSApi,
    remotepath: str,
    encoding: Optional[str] = None,
    encrypt_password: bytes = b"",
):
    pcs_file = api.path(remotepath)

    if not pcs_file:
        return

    fs = api.file_stream(pcs_file.file_id, encrypt_password=encrypt_password)
    if not fs:
        display_blocked_remotepath(remotepath)
        return

    cn = fs.read()
    if cn:
        if not encoding:
            r = chardet.detect(cn)
            if r["confidence"] > 0.5:
                encoding = r["encoding"]

        if encoding:
            print(cn.decode(encoding))
        else:
            print(cn)



from typing import Optional, Dict, Union
from pathlib import Path
import os
import mimetypes
import asyncio
import secrets
import copy
from urllib.parse import quote

import uvicorn

from alipcs_py.alipcs import AliPCSApi
from alipcs_py.common.io import RangeRequestIO
from alipcs_py.common.constant import CPU_NUM
from alipcs_py.common.path import join_path
from alipcs_py.utils import format_date

from fastapi import Depends, FastAPI, Request, status, HTTPException
from fastapi.security import HTTPBasic, HTTPBasicCredentials
from fastapi.responses import HTMLResponse, StreamingResponse, Response
from jinja2 import Template

from rich import print

mimetypes.init()

app = FastAPI()

_api: Optional[AliPCSApi] = None
_root_dir: str = "/"
_encrypt_password: bytes = b""

# For Auth
_username: Optional[str] = None
_password: Optional[str] = None

# This template is from https://github.com/rclone/rclone/blob/master/cmd/serve/httplib/serve/data/templates/index.html
_html_tempt: Template = Template((Path(__file__).parent / "index.html").open(encoding="utf-8").read())


def fake_io(io: RangeRequestIO, start: int = 0, end: int = -1):
    for b in io._auto_decrypt_request.read((start, end)):
        yield b


def get_file(file_id: str, remotepath: str, _range: Optional[str]):
    global _api
    assert _api

    try:
        fs = _api.file_stream(file_id, encrypt_password=_encrypt_password)
    except Exception as err:
        raise HTTPException(status_code=500, detail=f"Error: {err}, remotepath: {remotepath}")

    if not fs:
        raise HTTPException(status_code=404, detail=f"No download link: {remotepath}")

    length = len(fs)

    headers: Dict[str, str] = {
        "accept-ranges": "bytes",
        "connection": "Keep-Alive",
        "access-control-allow-origin": "*",
    }

    ext = os.path.splitext(remotepath)[-1]
    content_type = mimetypes.types_map.get(ext)

    if content_type:
        headers["content-type"] = content_type

    if _range and fs.seekable():
        assert _range.startswith("bytes=")

        status_code = 206
        start, end = _range[6:].split("-")
        _s, _e = int(start or 0), int(end or length - 1) + 1
        _iter_io = fake_io(fs, _s, _e)
        headers["content-range"] = f"bytes {_s}-{_e-1}/{length}"
        headers["content-length"] = str(_e - _s)
    else:
        status_code = 200
        _iter_io = fake_io(fs)
        headers["content-length"] = str(length)
    return StreamingResponse(_iter_io, status_code=status_code, headers=headers)


async def handle_request(
    request: Request,
    remotepath: str,
    order: str = "asc",  # desc , asc
    sort: str = "name",  # name, time, size
    file_id: str = "",
):
    desc = order == "desc"
    name = sort == "name"
    time = sort == "time"
    size = sort == "size"

    global _root_dir
    global _api
    assert _api

    _range = request.headers.get("range")

    if file_id:
        return get_file(file_id, remotepath, _range)

    remotepath = remotepath.strip("/")

    _rp = join_path(_root_dir, remotepath)

    # Anti path traversal attack
    if not _rp.startswith(_root_dir):
        raise HTTPException(status_code=404, detail="Item not found")

    rpf = _api.path(_rp)
    if not rpf:
        raise HTTPException(status_code=404, detail="Item not found")

    is_dir = rpf.is_dir
    if is_dir:
        chunks = ["/"] + (remotepath.split("/") if remotepath != "" else [])
        navigation = [(i - 1, "../" * (len(chunks) - i), name) for i, name in enumerate(chunks, 1)]
        pcs_files = _api.list_path_iter(_rp, desc=desc, name=name, time=time, size=size)
        entries = []
        for f in pcs_files:
            p = Path(f.path)
            entries.append(
                (
                    f.file_id,
                    f.is_dir,
                    p.name,
                    quote(p.name),
                    f.size,
                    format_date(f.updated_at or 0),
                )
            )
        cn = _html_tempt.render(root_dir=remotepath, navigation=navigation, entries=entries)
        return HTMLResponse(cn)
    else:
        return get_file(rpf.file_id, remotepath, _range)


_security = HTTPBasic()


def to_auth(credentials: HTTPBasicCredentials = Depends(_security)) -> str:
    correct_username = secrets.compare_digest(credentials.username, _username or "")
    correct_password = secrets.compare_digest(credentials.password, _password or "")
    if not (correct_username and correct_password):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect email or password",
            headers={"WWW-Authenticate": "Basic"},
        )
    return credentials.username


def make_auth_http_server(path: str = ""):
    @app.get("%s/{remotepath:path}" % path)
    async def auth_http_server(username: str = Depends(to_auth), response: Response = Depends(handle_request)):
        if username:
            return response

    @app.get("/__fileid__/")
    async def auth_file_id(username: str = Depends(to_auth), response: Response = Depends(handle_request)):
        if username:
            return response


def make_http_server(path: str = ""):
    @app.get("%s/{remotepath:path}" % path)
    async def http_server(response: Response = Depends(handle_request)):
        return response

    @app.get("/__fileid__/")
    async def file_id(response: Response = Depends(handle_request)):
        return response


def start_server(
    api: AliPCSApi,
    root_dir: str = "/",
    path: str = "",
    host: str = "localhost",
    port: int = 8000,
    workers: int = CPU_NUM,
    encrypt_password: bytes = b"",
    log_level: str = "info",
    username: Optional[str] = None,
    password: Optional[str] = None,
):
    """Create a http server on remote `root_dir`"""

    global _encrypt_password
    _encrypt_password = encrypt_password

    global _root_dir
    _root_dir = root_dir

    global _api
    if not _api:
        _api = api

    global _username
    if not _username:
        _username = username

    global _password
    if not _password:
        _password = password

    if path == "/" or not path:
        path = ""
    else:
        path = "/" + path.strip("/")

    print(f"[yellow]Server running on[/yellow] [b]http://{host}:{port}{path}/[/b]")

    if _username and _password:
        make_auth_http_server(path)
    else:
        make_http_server(path)

    log_config = copy.deepcopy(uvicorn.config.LOGGING_CONFIG)
    log_config["formatters"]["access"][
        "fmt"
    ] = '%(asctime)s - %(levelprefix)s %(client_addr)s - "%(request_line)s" %(status_code)s - %(msecs)d ms'
    uvicorn.run(
        "alipcs_py.commands.server:app",
        host=host,
        port=port,
        log_level=log_level,
        log_config=log_config,
        workers=1,
    )



from typing import Tuple
from io import StringIO
import time

import qrcode


_print = print

from rich import print

from alipcs_py.alipcs import AliOpenAuth, AliPCSError
from alipcs_py.alipcs.inner import AuthInfo


def openapi_qrcode_login(client_id: str, client_secret: str, client_server: str) -> AuthInfo:
    aliopenauth = AliOpenAuth(client_id=client_id, client_secret=client_secret, client_server=client_server)

    # Get qrcode info
    info = aliopenauth.get_qrcode_info()
    sid = info["sid"]

    qrcode_url = f"https://www.aliyundrive.com/o/oauth/authorize?sid={sid}"

    qr = qrcode.QRCode()
    qr.add_data(qrcode_url)
    f = StringIO()
    qr.print_ascii(out=f, tty=False, invert=True)
    f.seek(0)
    _print(f.read())

    print("  [red b]Please scan the qrcode to login in 120 seconds[/red b]")

    interval = 2 * 60  # wait 2min
    sleep = 2

    auth_code = ""
    for _ in range(interval // sleep):
        time.sleep(2)

        info = aliopenauth.scan_status(sid)
        if info["status"] == "LoginSuccess":
            auth_code = info["authCode"]
            break

    if not auth_code:
        raise AliPCSError("Login failed")

    info = aliopenauth.get_refresh_token(auth_code)
    return AuthInfo.from_(info)



from typing import Optional, List, Dict, Tuple, Union, final
import os


# Enable UTF-8 Mode for Windows
# https://www.python.org/dev/peps/pep-0540/
if os.name == "nt":
    os.environ["PYTHONUTF8"] = "1"

from collections import OrderedDict
from functools import wraps
from multiprocessing import Process
from pathlib import Path
import signal
import time
import logging
import traceback

from alipcs_py import __version__
from alipcs_py.alipcs import AliPCSApi, AliPCSError
from alipcs_py.app import account as account_module
from alipcs_py.app.account import Account, AccountManager
from alipcs_py.common.progress_bar import _progress, init_progress_bar
from alipcs_py.common.path import join_path
from alipcs_py.common.net import random_avail_port
from alipcs_py.common.io import EncryptType
from alipcs_py.common.event import keyboard_listener_start
from alipcs_py.commands.login import openapi_qrcode_login
from alipcs_py.commands.env import CONFIG_PATH, ACCOUNT_DATA_PATH
from alipcs_py.commands.sifter import (
    IncludeSifter,
    ExcludeSifter,
    IsFileSifter,
    IsDirSifter,
)
from alipcs_py.commands.display import (
    display_invalid_shared_link_infos,
    display_shared_files,
    display_shared_link_infos,
    display_user_info,
    display_user_infos,
)
from alipcs_py.commands.list_files import list_files
from alipcs_py.commands.cat import cat as _cat
from alipcs_py.commands import file_operators
from alipcs_py.commands.search import search as _search
from alipcs_py.commands.download import (
    download as _download,
    Downloader,
    DownloadParams,
    DEFAULT_DOWNLOADER,
    DEFAULT_CONCURRENCY,
    DEFAULT_CHUNK_SIZE,
)
from alipcs_py.commands.play import play as _play, Player, DEFAULT_PLAYER
from alipcs_py.commands.upload import upload as _upload, from_tos, CPU_NUM, UploadType
from alipcs_py.commands.sync import sync as _sync
from alipcs_py.commands import share as _share
from alipcs_py.commands.server import start_server
from alipcs_py.commands.log import get_logger
from alipcs_py.storage.store import AliPCSApiMixWithSharedStore, SharedStore
from alipcs_py.config import AppConfig
from alipcs_py.app.config import init_config

import click

from rich import print
from rich.console import Console
from rich.prompt import Prompt

logger = get_logger(__name__)

DEBUG = logger.level == logging.DEBUG


def _teardown():
    logger.debug("`app`: _teardown: start")

    from alipcs_py.common.event import _KEYBOARD_LISTENER

    if _KEYBOARD_LISTENER is not None:
        logger.debug("`app`: _teardown: reset old terminal")
        _KEYBOARD_LISTENER.set_normal_term()

    logger.debug("`app`: _teardown: end")


def _exit_progress_bar():
    if _progress.live._started:
        logger.debug("`app`: _exit_progress_bar: stop progress bar")

        print()
        # Stop _progress, otherwise terminal stdout will be contaminated
        _progress.stop()


def handle_signal(sign_num, frame):
    logger.debug("`app`: handle_signal: %s", sign_num)

    _exit_progress_bar()
    _teardown()

    # No use sys.exit() which only exits the main thread
    os._exit(1)


signal.signal(signal.SIGINT, handle_signal)


def handle_error(func):
    """Handle command error wrapper"""

    @wraps(func)
    def wrap(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except AliPCSError as err:
            _exit_progress_bar()

            logger.debug("`app`: AliPCSError: %s", traceback.format_exc())

            print(f"(v{__version__}) [bold red]ERROR[/bold red]: AliPCSError: {err}")
            if DEBUG:
                console = Console()
                console.print_exception()

        except Exception as err:
            _exit_progress_bar()

            logger.debug("`app`: System Error: %s", traceback.format_exc())

            print(f"(v{__version__}) [bold red]System ERROR[/bold red]: {err}")
            if DEBUG:
                console = Console()
                console.print_exception()

        finally:
            _exit_progress_bar()
            _teardown()

    return wrap


def save_modified_user_data(func):
    """Save modified user data after some commands is finished"""

    @wraps(func)
    def wrap(ctx, *args, **kwargs):
        pre_account_data_path_mtime = ctx.obj.pre_account_data_path_mtime

        try:
            result = func(ctx, *args, **kwargs)
            return result
        except:
            raise
        finally:
            now_account_data_path_mtime = Path(ctx.obj.account_data_path).stat().st_mtime
            # If the account_data had saved, we ignore to save it
            if now_account_data_path_mtime == pre_account_data_path_mtime:
                am = ctx.obj.account_manager
                used_user_ids = ctx.obj.used_user_ids
                for user_id in used_user_ids:
                    am.update(user_id)
                am.save()

    return wrap


def _app_config(ctx):
    """App Configuration"""

    return ctx.obj.app_config


def _user_ids(ctx) -> Optional[List[int]]:
    """Select use_ids by their name probes"""

    am = ctx.obj.account_manager
    account_name_probes = ctx.obj.accounts

    user_ids = []
    for user_id, account in am._accounts.items():
        account_name = account.account_name
        for probe in account_name_probes:
            if probe in account_name:
                user_ids.append(user_id)
                break
    return user_ids


def _change_account(ctx, user_id: int):
    """Change recent account with `user_id`"""

    am = ctx.obj.account_manager
    am.su(user_id)


def multi_user_do(func):
    """Run command on multi users"""

    @wraps(func)
    def wrap(*args, **kwargs):
        ctx = args[0]
        used_user_ids = ctx.obj.used_user_ids
        user_ids = _user_ids(ctx)
        if not user_ids:
            return func(*args, **kwargs)

        am = ctx.obj.account_manager
        for user_id in user_ids:
            accout = am.who(user_id)
            if not accout:
                continue

            user_name = accout.user.user_name
            print("[i yellow]@Do[/i yellow]: " f"user_name: [b]{user_name}[/b], " f"user_id: [b]{user_id}[/b]")
            _change_account(ctx, user_id)
            func(*args, **kwargs)
            print()

            used_user_ids.add(user_id)

    return wrap


def _recent_account(ctx) -> Optional[Account]:
    """Return recent user's `AliPCSApi`"""

    am = ctx.obj.account_manager
    account = am.who()
    if account:
        return account
    else:
        print("[italic red]No recent user, please adding or selecting one[/]")
        return None


def _recent_api(ctx) -> Union[AliPCSApi, AliPCSApiMixWithSharedStore, None]:
    """Return recent user's `AliPCSApi`"""

    am = ctx.obj.account_manager
    app_config = _app_config(ctx)
    account = _recent_account(ctx)
    if account:
        # Record used user_id for save_modified_user_data
        used_user_ids = ctx.obj.used_user_ids
        used_user_ids.add(account.user.user_id)

        api = am.get_api()
        if app_config.share.store:
            api._sharedstore = SharedStore()  # type: ignore
        return api
    else:
        return None


def _pwd(ctx) -> str:
    """Return recent user's pwd"""

    am = ctx.obj.account_manager
    return am.pwd


def _encrypt_password(ctx) -> bytes:
    """Return recent user's encryption key"""

    account = _recent_account(ctx)
    if account:
        ep = account.encrypt_password
        return bytes(ep or "", encoding="utf-8")
    else:
        return b""


ALIAS = OrderedDict(
    **{
        # Account
        "w": "who",
        "uu": "updateuser",
        "su": "su",
        "ul": "userlist",
        "ua": "useradd",
        "ud": "userdel",
        "ep": "encryptpwd",
        # File Operations
        "l": "ls",
        "f": "search",
        "md": "mkdir",
        "mv": "move",
        "rn": "rename",
        "cp": "copy",
        "rm": "remove",
        "d": "download",
        "p": "play",
        "u": "upload",
        "sn": "sync",
        # Share
        "S": "share",
        "sl": "shared",
        "cs": "cancelshared",
        "s": "save",
        "ssl": "storesharedlinks",
        "lsl": "listsharedlinks",
        "lsf": "listsharedfiles",
        "fsl": "findsharedlinks",
        "fsf": "findsharedfiles",
        "fs": "findshared",
        "dss": "deletestoredshared",
        "cst": "cleanstore",
        # Server
        "sv": "server",
    }
)


class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        # As normal command name
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv

        # Check alias command name
        if cmd_name not in ALIAS:
            ctx.fail(f"No command: {cmd_name}")

        normal_cmd_name = ALIAS[cmd_name]
        return click.Group.get_command(self, ctx, normal_cmd_name)

    def list_commands(self, ctx):
        return self.commands


_APP_DOC = f"""AliPCS App v{__version__}

    \b
    如果第一次使用，你需要运行 `AliPCS-Py useradd` 添加 `refresh_token`。
    如何获取 `refresh_token` 见 https://github.com/PeterDing/AliPCS-Py#%E6%B7%BB%E5%8A%A0%E7%94%A8%E6%88%B7
    用 `AliPCS-Py {{command}} --help` 查看具体的用法。"""

_ALIAS_DOC = "Command 别名:\n\n\b\n" + "\n".join([f"{alias: >3} : {cmd}" for alias, cmd in ALIAS.items()])


@click.group(cls=AliasedGroup, help=_APP_DOC, epilog=_ALIAS_DOC)
@click.option("--config", "-c", type=str, default=CONFIG_PATH, help="Configuration file")
@click.option(
    "--account-data-path",
    "--adp",
    type=str,
    default=ACCOUNT_DATA_PATH,
    help="Account data file",
)
@click.option("--accounts", "-u", type=str, default=None, help="帐号名片段，用“,”分割")
@click.pass_context
def app(ctx, config, account_data_path, accounts):
    ctx.obj.account_data_path = account_data_path
    ad_path = Path(account_data_path)
    ctx.obj.pre_account_data_path_mtime = 0
    if ad_path.exists():
        ctx.obj.pre_account_data_path_mtime = ad_path.stat().st_mtime
    ctx.obj.account_manager = AccountManager.load_data(account_data_path)
    ctx.obj.accounts = [] if accounts is None else accounts.split(",")
    ctx.obj.used_user_ids = set()

    # Load app config
    app_config = AppConfig.load(config)
    ctx.obj.app_config = app_config
    init_config(app_config)

    if app_config.share.store:
        account_module.AliPCSApiMix = AliPCSApiMixWithSharedStore


# Account
# {{{
@app.command()
@click.argument("user_id", type=int, default=None, required=False)
@click.option("--show-encrypt-password", "-K", is_flag=True, help="显示加密密码")
@click.pass_context
@handle_error
def who(ctx, user_id, show_encrypt_password):
    """显示当前用户的信息

    也可指定 `user_id`
    """

    am = ctx.obj.account_manager
    account = am.who(user_id)
    if account:
        display_user_info(account.user)
        if show_encrypt_password:
            encrypt_password = _encrypt_password(ctx)

            print(f"[red]encrypt password[/red]: {encrypt_password}")
    else:
        print("[italic red]No recent user, please adding or selecting one[/]")


@app.command()
@click.argument("user_ids", type=int, nargs=-1, default=None, required=False)
@click.pass_context
@handle_error
def updateuser(ctx, user_ids):
    """更新用户信息 （默认更新当前用户信息）

    也可指定多个 `user_id`
    """

    am = ctx.obj.account_manager
    if not user_ids:
        user_ids = [am._who]

    for user_id in user_ids:
        am.update(user_id)
        account = am.who(user_id)
        if account:
            display_user_info(account.user)

    am.save()


@app.command()
@click.argument("user_index", type=int, default=None, required=False)
@click.pass_context
@handle_error
def su(ctx, user_index):
    """切换当前用户"""

    am = ctx.obj.account_manager
    ls = sorted(
        [(a.user, a.pwd, a.account_name) for a in am.accounts],
        key=lambda x: x[0].user_id,
    )
    display_user_infos(*ls, recent_user_id=am._who)

    if user_index:
        assert 0 < user_index < len(ls) + 1, f"Out range index {user_index}"

        i = user_index
        print(f"Select the user index: {i}")
    else:
        indexes = list(str(idx) for idx in range(1, len(ls) + 1))
        i = Prompt.ask("Select an user index", choices=indexes, default="")

        if not i:
            return

    user_id = ls[int(i) - 1][0].user_id
    am.su(user_id)
    am.save()


@app.command()
@click.pass_context
@handle_error
def userlist(ctx):
    """显示所有用户"""

    am = ctx.obj.account_manager
    ls = sorted(
        [(a.user, a.pwd, a.account_name) for a in am.accounts],
        key=lambda x: x[0].user_id,
    )
    display_user_infos(*ls, recent_user_id=am._who)


@app.command()
@click.option("--account-name", prompt="Account Name", hide_input=False, default="", help="账号名 [默认为 user id]")
@click.option("--web-refresh-token", prompt="web refresh token", hide_input=True, help="用户 web_refresh_token")
@click.option(
    "--openapi-refresh-token",
    prompt="openapi refresh token",
    hide_input=True,
    default="",
    help="用户 openapi_refresh_token",
)
@click.option("--client-id", prompt="open client id", default="", help="openapi client id")
@click.option("--client-secret", prompt="open client secret", default="", help="openapi client secret")
@click.option("--client-server", prompt="open client server", default="", help="openapi client server")
@click.pass_context
@handle_error
def useradd(ctx, account_name, web_refresh_token, openapi_refresh_token, client_id, client_secret, client_server):
    """添加一个用户并设置为当前用户"""

    assert web_refresh_token, "No `web_refresh_token`"

    openapi_access_token = ""
    openapi_token_type = "Bearer"
    openapi_expire_time = 0
    if (client_id and client_secret) or client_server:
        if not openapi_refresh_token:
            auth_info = openapi_qrcode_login(client_id, client_secret, client_server)
            openapi_refresh_token = auth_info.refresh_token
            openapi_access_token = auth_info.access_token
            openapi_token_type = auth_info.token_type
            openapi_expire_time = auth_info.expire_time

    account = Account.from_refresh_token(
        web_refresh_token,
        openapi_refresh_token=openapi_refresh_token,
        openapi_access_token=openapi_access_token,
        openapi_token_type=openapi_token_type,
        openapi_expire_time=openapi_expire_time,
        client_id=client_id or "",
        client_secret=client_secret or "",
        client_server=client_server or "",
        account_name=account_name,
    )
    am = ctx.obj.account_manager
    am.add_account(account)
    am.su(account.user.user_id)
    am.save()


@app.command()
@click.pass_context
@handle_error
def userdel(ctx):
    """删除一个用户"""

    am = ctx.obj.account_manager
    ls = sorted(
        [(a.user, a.pwd, a.account_name) for a in am.accounts],
        key=lambda x: x[0].user_id,
    )
    display_user_infos(*ls, recent_user_id=am._who)

    indexes = list(str(idx) for idx in range(1, len(ls) + 1))
    i = Prompt.ask("Delete an user index", choices=indexes, default="")
    if not i:
        return

    user_id = ls[int(i) - 1][0].user_id
    am.delete_account(user_id)
    am.save()

    print(f"Delete user {user_id}")


@app.command()
@click.option(
    "--encrypt-password",
    "--ep",
    prompt="encrypt-password",
    hide_input=True,
    help="加密密码，任意字符",
)
@click.pass_context
@handle_error
def encryptpwd(ctx, encrypt_password):
    """设置加密密码"""

    assert len(encrypt_password) > 0, "No encrypt-password"

    am = ctx.obj.account_manager
    am.set_encrypt_password(encrypt_password)
    am.save()


@app.command()
@click.argument("remotedir", type=str, default="/", required=False)
@click.pass_context
@handle_error
def cd(ctx, remotedir):
    """切换当前工作目录"""

    am = ctx.obj.account_manager
    am.cd(remotedir)
    am.save()


@app.command()
@click.pass_context
@handle_error
@multi_user_do
def pwd(ctx):
    """显示当前工作目录"""

    pwd = _pwd(ctx)
    print(pwd)


# }}}


# File Operations
# {{{
@app.command()
@click.argument("remotepaths", nargs=-1, type=str)
@click.option("--file-id", "-i", multiple=True, type=str, help="文件 ID")
@click.option("--share-id", "--si", nargs=1, type=str, help="列出这个分享ID下的文件")
@click.option("--share-url", "--su", nargs=1, type=str, help="列出这个分享url下的文件")
@click.option("--password", "-p", type=str, help="分享链接密码，如果没有不用设置")
@click.option("--desc", "-r", is_flag=True, help="逆序排列文件")
@click.option("--name", "-n", is_flag=True, help="依名字排序")
@click.option("--time", "-t", is_flag=True, help="依时间排序")
@click.option("--size", "-s", is_flag=True, help="依文件大小排序")
@click.option("--recursive", "-R", is_flag=True, help="递归列出文件")
@click.option("--include", "-I", type=str, help="筛选包含这个字符串的文件")
@click.option("--include-regex", "--IR", type=str, help="筛选包含这个正则表达式的文件")
@click.option("--exclude", "-E", type=str, help="筛选 不 包含这个字符串的文件")
@click.option("--exclude-regex", "--ER", type=str, help="筛选 不 包含这个正则表达式的文件")
@click.option("--is-file", "-f", is_flag=True, help="筛选 非 目录文件")
@click.option("--is-dir", "-d", is_flag=True, help="筛选目录文件")
@click.option("--no-highlight", "--NH", is_flag=True, help="取消匹配高亮")
@click.option("--show-size", "-S", is_flag=True, help="显示文件大小")
@click.option("--show-date", "-D", is_flag=True, help="显示文件创建时间")
@click.option("--show-file-id", "--ID", is_flag=True, help="显示文件 ID")
@click.option("--show-hash", "-H", is_flag=True, help="显示文件 sha1")
@click.option("--show-absolute-path", "-A", is_flag=True, help="显示文件绝对路径")
@click.option("--show-dl-link", "--DL", is_flag=True, help="显示文件下载连接")
@click.option("--csv", is_flag=True, help="用 csv 格式显示，单行显示，推荐和 --DL 或 --HL 一起用")
@click.option("--only-dl-link", "--ODL", is_flag=True, help="只显示文件下载连接")
@click.pass_context
@handle_error
@save_modified_user_data
@multi_user_do
def ls(
    ctx,
    remotepaths,
    file_id,
    share_id,
    share_url,
    password,
    desc,
    name,
    time,
    size,
    recursive,
    include,
    include_regex,
    exclude,
    exclude_regex,
    is_file,
    is_dir,
    no_highlight,
    show_size,
    show_date,
    show_file_id,
    show_hash,
    show_absolute_path,
    show_dl_link,
    csv,
    only_dl_link,
):
    """列出网盘路径下的文件和对应的文件信息"""

    api = _recent_api(ctx)
    if not api:
        return

    sifters = []
    if include:
        sifters.append(IncludeSifter(include, regex=False))
    if include_regex:
        sifters.append(IncludeSifter(include_regex, regex=True))
    if exclude:
        sifters.append(ExcludeSifter(exclude, regex=False))
    if exclude_regex:
        sifters.append(ExcludeSifter(exclude_regex, regex=True))
    if is_file:
        sifters.append(IsFileSifter())
    if is_dir:
        sifters.append(IsDirSifter())

    if share_id or share_url:
        if not file_id and not remotepaths and "folder" not in (share_url or ""):
            remotepaths = ["/"]

        assert all([r.startswith("/") for r in remotepaths])

        _share.list_shared_files(
            api,
            *remotepaths,
            share_id=share_id,
            share_url=share_url,
            password=password,
            file_ids=file_id,
            desc=desc,
            name=name,
            time=time,
            size=size,
            recursive=recursive,
            sifters=sifters,
            highlight=not no_highlight,
            show_size=show_size,
            show_date=show_date,
            show_file_id=show_file_id,
            show_hash=show_hash,
            show_absolute_path=show_absolute_path,
            csv=csv,
        )
    else:
        pwd = _pwd(ctx)
        if not file_id and not remotepaths:
            remotepaths = [pwd]

        remotepaths = [join_path(pwd, r) for r in list(remotepaths)]

        list_files(
            api,
            *remotepaths,
            file_ids=file_id,
            desc=desc,
            name=name,
            time=time,
            size=size,
            recursive=recursive,
            sifters=sifters,
            highlight=not no_highlight,
            show_size=show_size,
            show_date=show_date,
            show_file_id=show_file_id,
            show_hash=show_hash,
            show_absolute_path=show_absolute_path,
            show_dl_link=show_dl_link,
            csv=csv,
            only_dl_link=only_dl_link,
        )


@app.command()
@click.argument("keyword", nargs=1, type=str)
@click.option("--include", "-I", type=str, help="筛选包含这个字符串的文件")
@click.option("--include-regex", "--IR", type=str, help="筛选包含这个正则表达式的文件")
@click.option("--exclude", "-E", type=str, help="筛选 不 包含这个字符串的文件")
@click.option("--exclude-regex", "--ER", type=str, help="筛选 不 包含这个正则表达式的文件")
@click.option("--is-file", "-f", is_flag=True, help="筛选 非 目录文件")
@click.option("--is-dir", "-d", is_flag=True, help="筛选目录文件")
@click.option("--no-highlight", "--NH", is_flag=True, help="取消匹配高亮")
@click.option("--show-size", "-S", is_flag=True, help="显示文件大小")
@click.option("--show-date", "-D", is_flag=True, help="显示文件创建时间")
@click.option("--show-hash", "-H", is_flag=True, help="显示文件 sha1")
@click.option("--csv", is_flag=True, help="用 csv 格式显示")
@click.pass_context
@handle_error
@save_modified_user_data
@multi_user_do
def search(
    ctx,
    keyword,
    include,
    include_regex,
    exclude,
    exclude_regex,
    is_file,
    is_dir,
    no_highlight,
    show_size,
    show_date,
    show_hash,
    csv,
):
    """搜索包含 `keyword` 的文件"""

    api = _recent_api(ctx)
    if not api:
        return

    sifters = []
    if include:
        sifters.append(IncludeSifter(include, regex=False))
    if include_regex:
        sifters.append(IncludeSifter(include_regex, regex=True))
    if exclude:
        sifters.append(ExcludeSifter(exclude, regex=False))
    if exclude_regex:
        sifters.append(ExcludeSifter(exclude_regex, regex=True))
    if is_file:
        sifters.append(IsFileSifter())
    if is_dir:
        sifters.append(IsDirSifter())

    _search(
        api,
        keyword,
        sifters=sifters,
        highlight=not no_highlight,
        show_size=show_size,
        show_date=show_date,
        show_file_id=True,
        show_hash=show_hash,
        csv=csv,
    )


@app.command()
@click.argument("remotepath", nargs=1, type=str)
@click.option("--encoding", "-e", type=str, help="文件编码，默认自动解码")
@click.option("--no-decrypt", "--ND", is_flag=True, help="不解密")
@click.option("--encrypt-password", "--ep", type=str, default=None, help="加密密码，默认使用用户设置的")
@click.pass_context
@handle_error
@save_modified_user_data
@multi_user_do
def cat(ctx, remotepath, encoding, no_decrypt, encrypt_password):
    """显示文件内容"""

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    remotepath = join_path(pwd, remotepath)

    if no_decrypt:
        encrypt_password = b""
    else:
        encrypt_password = encrypt_password or _encrypt_password(ctx)

    _cat(api, remotepath, encoding=encoding, encrypt_password=encrypt_password)


@app.command()
@click.argument("remotedirs", nargs=-1, type=str)
@click.option("--show", "-S", is_flag=True, help="显示目录")
@click.pass_context
@handle_error
@save_modified_user_data
@multi_user_do
def mkdir(ctx, remotedirs, show):
    """创建目录"""

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    remotedirs = (join_path(pwd, d) for d in remotedirs)

    file_operators.makedir(api, *remotedirs, show=show)


@app.command()
@click.argument("remotepaths", nargs=-1, type=str)
@click.option("--show", "-S", is_flag=True, help="显示结果")
@click.pass_context
@handle_error
@save_modified_user_data
@multi_user_do
def move(ctx, remotepaths, show):
    """移动文件

    \b
    examples:
        move /file1 /file2 /to/dir
    """

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    remotepaths = [join_path(pwd, r) for r in remotepaths]

    if len(remotepaths) < 2:
        ctx.fail("remote paths < 2")
    file_operators.move(api, *remotepaths, show=show)


@app.command()
@click.argument("remotepath", nargs=1, type=str)
@click.argument("new_name", nargs=1, type=str)
@click.option("--show", "-S", is_flag=True, help="显示结果")
@click.pass_context
@handle_error
@save_modified_user_data
@multi_user_do
def rename(ctx, remotepath, new_name, show):
    """文件重命名

    \b
    examples:
        rename /path/to/far new_name_foo
        # /path/to/far -> /path/to/new_name_foo
    """

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    remotepath = join_path(pwd, remotepath)

    file_operators.rename(api, remotepath, new_name, show=show)


@app.command()
@click.argument("remotepaths", nargs=-1, type=str)
@click.option("--show", "-S", is_flag=True, help="显示结果")
@click.pass_context
@handle_error
@save_modified_user_data
@multi_user_do
def copy(ctx, remotepaths, show):
    """拷贝文件

    \b
    examples:
        copy /file1 /file2 /to/dir
    """

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    remotepaths = [join_path(pwd, r) for r in remotepaths]

    if len(remotepaths) < 2:
        ctx.fail("remote paths < 2")
    file_operators.copy(api, *remotepaths, show=show)


@app.command()
@click.argument("remotepaths", nargs=-1, type=str)
@click.option("--file-id", "-i", multiple=True, type=str, help="文件 ID")
@click.pass_context
@handle_error
@save_modified_user_data
@multi_user_do
def remove(ctx, remotepaths, file_id):
    """删除文件"""

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    remotepaths = (join_path(pwd, r) for r in remotepaths)

    file_operators.remove(api, *remotepaths)

    if file_id:
        api.remove(*file_id)


@app.command()
@click.argument("remotepaths", nargs=-1, type=str)
@click.option("--file-id", "-i", multiple=True, type=str, help="文件 ID")
@click.option("--outdir", "-o", nargs=1, type=str, default=".", help="指定下载本地目录，默认为当前目录")
@click.option("--share-id", "--si", nargs=1, type=str, help="下载这个分享ID下的文件")
@click.option("--share-url", "--su", nargs=1, type=str, help="下载这个分享url下的文件")
@click.option("--password", "-p", type=str, help="分享链接密码，如果没有不用设置")
@click.option("--recursive", "-R", is_flag=True, help="递归下载")
@click.option("--from-index", "-f", type=int, default=0, help="从所有目录中的第几个文件开始下载，默认为0（第一个）")
@click.option("--include", "-I", type=str, help="筛选包含这个字符串的文件")
@click.option("--include-regex", "--IR", type=str, help="筛选包含这个正则表达式的文件")
@click.option("--exclude", "-E", type=str, help="筛选 不 包含这个字符串的文件")
@click.option("--exclude-regex", "--ER", type=str, help="筛选 不 包含这个正则表达式的文件")
@click.option(
    "-d",
    "--downloader",
    type=click.Choice([d.name for d in Downloader]),
    default=DEFAULT_DOWNLOADER.name,
    help="""指定下载应用

    \b
    默认为 me (AliPCS-Py 自己的下载器，支持断续下载)
        me 使用多文件并发下载。

    除 me 外，其他下载器，不使用多文件并发下载，使用一个文件多链接下载。

    \b
    aget_py (https://github.com/PeterDing/aget) 默认安装
    aget_rs (下载 https://github.com/PeterDing/aget-rs/releases)
    aria2 (下载 https://github.com/aria2/aria2/releases)
    """,
)
@click.option(
    "--concurrency",
    "-s",
    type=int,
    default=DEFAULT_CONCURRENCY,
    help=f"下载同步链接数，默认为{DEFAULT_CONCURRENCY}。建议小于 10",
)
@click.option("--chunk-size", "-k", type=str, default=DEFAULT_CHUNK_SIZE, help="同步链接分块大小")
@click.option("--no-decrypt", "--ND", is_flag=True, help="不解密")
@click.option("--quiet", "-q", is_flag=True, help="取消第三方下载应用输出")
@click.option("--out-cmd", "--OC", is_flag=True, help="输出第三方下载应用命令")
@click.option("--encrypt-password", "--ep", type=str, default=None, help="加密密码，默认使用用户设置的")
@click.pass_context
@handle_error
@save_modified_user_data
@multi_user_do
def download(
    ctx,
    remotepaths,
    file_id,
    outdir,
    share_id,
    share_url,
    password,
    recursive,
    from_index,
    include,
    include_regex,
    exclude,
    exclude_regex,
    downloader,
    concurrency,
    chunk_size,
    no_decrypt,
    quiet,
    out_cmd,
    encrypt_password,
):
    """下载文件"""

    if out_cmd:
        assert downloader != Downloader.me.name, "输出命令只能用于第三方下载应用"

    api = _recent_api(ctx)
    if not api:
        return

    sifters = []
    if include:
        sifters.append(IncludeSifter(include, regex=False))
    if include_regex:
        sifters.append(IncludeSifter(include_regex, regex=True))
    if exclude:
        sifters.append(ExcludeSifter(exclude, regex=False))
    if exclude_regex:
        sifters.append(ExcludeSifter(exclude_regex, regex=True))

    if no_decrypt:
        encrypt_password = b""
    else:
        encrypt_password = encrypt_password or _encrypt_password(ctx)

    if not quiet and downloader == Downloader.me.name:
        init_progress_bar()

    if share_id or share_url:
        assert all([r.startswith("/") for r in remotepaths])

        _share.download_shared(
            api,
            remotepaths,
            file_id,
            outdir,
            share_id=share_id,
            share_url=share_url,
            password=password,
            sifters=sifters,
            recursive=recursive,
            from_index=from_index,
            downloader=getattr(Downloader, downloader),
            downloadparams=DownloadParams(concurrency=concurrency, chunk_size=chunk_size, quiet=quiet),
            out_cmd=out_cmd,
            encrypt_password=encrypt_password,
        )
    else:
        pwd = _pwd(ctx)
        remotepaths = [join_path(pwd, r) for r in remotepaths]
        _download(
            api,
            remotepaths,
            file_id,
            outdir,
            sifters=sifters,
            recursive=recursive,
            from_index=from_index,
            downloader=getattr(Downloader, downloader),
            downloadparams=DownloadParams(concurrency=concurrency, chunk_size=chunk_size, quiet=quiet),
            out_cmd=out_cmd,
            encrypt_password=encrypt_password,
        )


@app.command()
@click.argument("remotepaths", nargs=-1, type=str)
@click.option("--file-id", "-i", multiple=True, type=str, help="文件 ID")
@click.option("--share-id", "--si", nargs=1, type=str, help="播放这个分享ID下的文件")
@click.option("--share-url", "--su", nargs=1, type=str, help="播放这个分享url下的文件")
@click.option("--password", "-p", type=str, help="分享链接密码，如果没有不用设置")
@click.option("--recursive", "-R", is_flag=True, help="递归播放")
@click.option("--from-index", "-f", type=int, default=0, help="从所有目录中的第几个文件开始播放，默认为0（第一个）")
@click.option("--include", "-I", type=str, help="筛选包含这个字符串的文件")
@click.option("--include-regex", "--IR", type=str, help="筛选包含这个正则表达式的文件")
@click.option("--exclude", "-E", type=str, help="筛选 不 包含这个字符串的文件")
@click.option("--exclude-regex", "--ER", type=str, help="筛选 不 包含这个正则表达式的文件")
@click.option(
    "--player",
    "--pl",
    type=click.Choice([d.name for d in Player]),
    default=DEFAULT_PLAYER.name,
    help="""指定第三方播放器

    \b
    默认为 mpv (https://mpv.io),
    """,
)
@click.option("--player-params", "--PP", multiple=True, type=str, help="第三方播放器参数")
@click.option("--quiet", "-q", is_flag=True, help="取消第三方播放器输出")
@click.option("--shuffle", "--sf", is_flag=True, help="随机播放")
@click.option("--ignore-ext", "--IE", is_flag=True, help="不用文件名后缀名来判断媒体文件")
@click.option("--out-cmd", "--OC", is_flag=True, help="输出第三方播放器命令")
@click.option("--use-local-server", "-s", is_flag=True, help="使用本地服务器播放。")
@click.option("--encrypt-password", "--ep", type=str, default=None, help="加密密码，默认使用用户设置的")
@click.pass_context
@handle_error
@save_modified_user_data
@multi_user_do
def play(
    ctx,
    remotepaths,
    file_id,
    share_id,
    share_url,
    password,
    recursive,
    from_index,
    include,
    include_regex,
    exclude,
    exclude_regex,
    player,
    player_params,
    quiet,
    shuffle,
    ignore_ext,
    out_cmd,
    use_local_server,
    encrypt_password,
):
    """播放媒体文件"""

    api = _recent_api(ctx)
    if not api:
        return

    sifters = []
    if include:
        sifters.append(IncludeSifter(include, regex=False))
    if include_regex:
        sifters.append(IncludeSifter(include_regex, regex=True))
    if exclude:
        sifters.append(ExcludeSifter(exclude, regex=False))
    if exclude_regex:
        sifters.append(ExcludeSifter(exclude_regex, regex=True))

    local_server = ""
    if use_local_server:
        if share_id or file_id:
            assert ValueError("Recently local server can't play others shared items and using `file_id`")

        encrypt_password = encrypt_password or _encrypt_password(ctx)

        host = "localhost"
        port = random_avail_port()

        local_server = f"http://{host}:{port}"

        ps = Process(
            target=start_server,
            args=(
                api,
                "/",
            ),
            kwargs=dict(
                host=host,
                port=port,
                workers=CPU_NUM,
                encrypt_password=encrypt_password,
                log_level="warning",
            ),
        )
        ps.start()
        time.sleep(1)

    if share_id or share_url:
        assert all([r.startswith("/") for r in remotepaths])

        _share.play_shared(
            api,
            remotepaths,
            file_ids=file_id,
            share_id=share_id,
            share_url=share_url,
            password=password,
            sifters=sifters,
            recursive=recursive,
            from_index=from_index,
            player=getattr(Player, player),
            player_params=player_params,
            quiet=quiet,
            shuffle=shuffle,
            ignore_ext=ignore_ext,
            out_cmd=out_cmd,
            local_server=local_server,
        )
    else:
        pwd = _pwd(ctx)
        remotepaths = [join_path(pwd, r) for r in list(remotepaths)]
        _play(
            api,
            remotepaths,
            file_ids=file_id,
            sifters=sifters,
            recursive=recursive,
            from_index=from_index,
            player=getattr(Player, player),
            player_params=player_params,
            quiet=quiet,
            shuffle=shuffle,
            ignore_ext=ignore_ext,
            out_cmd=out_cmd,
            local_server=local_server,
        )

    if use_local_server:
        ps.terminate()


@app.command()
@click.argument("localpaths", nargs=-1, type=str)
@click.argument("remotedir", nargs=1, type=str)
@click.option(
    "--upload-type",
    "-t",
    type=click.Choice([t.name for t in UploadType]),
    default=UploadType.Many.name,
    help=("上传方式，Many: 同时上传多个文件，" "One: 一次只上传一个文件，但同时上传文件的多个分片 " "(阿里网盘不支持单文件并发上传。`upload --upload-type One` 失效)"),
)
@click.option("--encrypt-password", "--ep", type=str, default=None, help="加密密码，默认使用用户设置的")
@click.option(
    "--encrypt-type",
    "-e",
    type=click.Choice([t.name for t in EncryptType]),
    default=EncryptType.No.name,
    help="文件加密方法，默认为 No 不加密",
)
@click.option("--max-workers", "-w", type=int, default=CPU_NUM, help="同时上传连接数量，默认为 CPU 核数")
@click.option("--no-ignore-existing", "--NI", is_flag=True, help="上传已经存在的文件")
@click.option("--no-show-progress", "--NP", is_flag=True, help="不显示上传进度")
@click.pass_context
@handle_error
@save_modified_user_data
@multi_user_do
def upload(
    ctx,
    localpaths,
    remotedir,
    upload_type,
    encrypt_password,
    encrypt_type,
    max_workers,
    no_ignore_existing,
    no_show_progress,
):
    """上传文件"""

    if upload_type == UploadType.One.name:
        raise ValueError("阿里网盘不支持单文件并发上传。`upload --upload-type One` 失效")

    # Keyboard listener start
    keyboard_listener_start()

    api = _recent_api(ctx)
    if not api:
        return

    encrypt_password = encrypt_password or _encrypt_password(ctx)
    if encrypt_type != EncryptType.No.name and not encrypt_password:
        raise ValueError(f"Encrypting with {encrypt_type} must have a key")

    pwd = _pwd(ctx)
    remotedir = join_path(pwd, remotedir)

    from_to_list = from_tos(localpaths, remotedir)

    if not no_show_progress:
        init_progress_bar()

    check_name_mode = "refuse" if not no_ignore_existing else "auto_rename"

    _upload(
        api,
        from_to_list,
        upload_type=getattr(UploadType, upload_type),
        check_name_mode=check_name_mode,
        encrypt_password=encrypt_password,
        encrypt_type=getattr(EncryptType, encrypt_type),
        max_workers=max_workers,
        show_progress=not no_show_progress,
    )


@app.command()
@click.argument("localdir", nargs=1, type=str)
@click.argument("remotedir", nargs=1, type=str)
@click.option("--encrypt-password", "--ep", type=str, default=None, help="加密密码，默认使用用户设置的")
@click.option(
    "--encrypt-type",
    "-e",
    type=click.Choice([t.name for t in EncryptType]),
    default=EncryptType.No.name,
    help="文件加密方法，默认为 No 不加密",
)
@click.option("--max-workers", "-w", type=int, default=CPU_NUM, help="同时上传文件数")
@click.option("--no-show-progress", "--NP", is_flag=True, help="不显示上传进度")
@click.pass_context
@handle_error
@save_modified_user_data
@multi_user_do
def sync(
    ctx,
    localdir,
    remotedir,
    encrypt_password,
    encrypt_type,
    max_workers,
    no_show_progress,
):
    """同步本地目录到远端"""

    # Keyboard listener start
    keyboard_listener_start()

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    remotedir = join_path(pwd, remotedir)

    encrypt_password = encrypt_password or _encrypt_password(ctx)
    if encrypt_type != EncryptType.No.name and not encrypt_password:
        raise ValueError(f"Encrypting with {encrypt_type} must have a key")

    _sync(
        api,
        localdir,
        remotedir,
        encrypt_password=encrypt_password,
        encrypt_type=getattr(EncryptType, encrypt_type),
        max_workers=max_workers,
        show_progress=not no_show_progress,
    )


# }}}


# Share
# {{{
@app.command()
@click.argument("remotepaths", nargs=-1, type=str)
@click.option("--password", "-p", type=str, default="", help="设置秘密，默认没有秘密")
@click.option("--period-time", "--pt", type=int, default=0, help="设置分享有效期，单位为天")
@click.pass_context
@handle_error
@save_modified_user_data
@multi_user_do
def share(ctx, remotepaths, password, period_time):
    """分享文件

    \b
    examples:
        share /path1 path2
    """

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    remotepaths = (join_path(pwd, r) for r in remotepaths)

    _share.share_files(api, *remotepaths, password=password, period=period_time or 0)


@app.command()
@click.option("--show-all", "-A", is_flag=True, help="显示所有分享的链接，默认只显示有效的分享链接")
@click.pass_context
@handle_error
@save_modified_user_data
@multi_user_do
def shared(ctx, show_all):
    """列出分享链接"""

    api = _recent_api(ctx)
    if not api:
        return

    _share.list_shared(api, show_all=show_all)


@app.command()
@click.argument("share_ids", nargs=-1, type=str)
@click.pass_context
@handle_error
@save_modified_user_data
@multi_user_do
def cancelshared(ctx, share_ids):
    """取消分享链接"""

    api = _recent_api(ctx)
    if not api:
        return

    _share.cancel_shared(api, *share_ids)


@app.command()
@click.argument("share_url_or_id", nargs=1, type=str)
@click.argument("remotedir", nargs=1, type=str)
@click.option("--file-id", "-i", multiple=True, type=str, help="文件 ID")
@click.option("--password", "-p", type=str, help="分享链接密码，如果没有不用设置")
@click.pass_context
@handle_error
@save_modified_user_data
@multi_user_do
def save(ctx, share_url_or_id, file_id, remotedir, password):
    """保存其他用户分享的链接"""

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    remotedir = join_path(pwd, remotedir)

    share_url = ""
    share_id = ""
    if "/s/" in share_url_or_id:
        share_url = share_url_or_id
    else:
        share_id = share_url_or_id

    _share.save_shared(
        api,
        remotedir,
        share_id=share_id,
        share_url=share_url,
        file_ids=file_id,
        password=password,
    )


@app.command()
@click.argument("share_urls_or_ids", nargs=-1, type=str)
@click.option("--password", "-p", type=str, help="分享链接密码，如果没有不用设置")
@click.pass_context
@handle_error
@save_modified_user_data
def storesharedlinks(ctx, share_urls_or_ids, password):
    """保存分享连接至本地

    注意: 使用这个命令必须将配置文件 ~/.alipcs-py/config.toml 中的 [share] store 设为 true
    """

    app_config = _app_config(ctx)
    if not app_config.share.store:
        print("App configuration `[share] store is false`. So the command does not work")
        return

    api = _recent_api(ctx)
    if not api:
        return

    for share_url_or_id in share_urls_or_ids:
        share_url = ""
        share_id = ""
        if "/s/" in share_url_or_id:
            share_url = share_url_or_id
        else:
            share_id = share_url_or_id

        _share.get_share_token(api, share_id=share_id, share_url=share_url, password=password)


@app.command()
@click.pass_context
@handle_error
def listsharedlinks(ctx):
    """显示本地保存的分享连接

    注意: 使用这个命令必须将配置文件 ~/.alipcs-py/config.toml 中的 [share] store 设为 true
    """

    shared_store = SharedStore()

    offset = 0
    while True:
        shared_links = shared_store.list_shared_links(offset=offset)
        display_shared_link_infos(*shared_links)
        yes = Prompt.ask("Next page", choices=["y", "n"], default="y")
        if yes == "y":
            offset += len(shared_links)
        else:
            break


@app.command()
@click.option("--share-id", "--si", multiple=True, type=str, help="指定显示 share id 下的文件")
@click.pass_context
@handle_error
def listsharedfiles(ctx, share_id):
    """显示本地保存的分享文件

    注意: 使用这个命令必须将配置文件 ~/.alipcs-py/config.toml 中的 [share] store 设为 true
    """

    shared_store = SharedStore()

    offset = 0
    while True:
        shared_files = shared_store.list_shared_files(share_ids=share_id, offset=offset)
        display_shared_files(*shared_files)
        yes = Prompt.ask("Next page", choices=["y", "n"], default="y")
        if yes == "y":
            offset += len(shared_files)
        else:
            break


def _find_shared_links(keywords: List[str]):
    shared_store = SharedStore()
    shared_links = shared_store.search_shared_links(*keywords)
    display_shared_link_infos(*shared_links)


@app.command()
@click.argument("keywords", nargs=-1, type=str)
@click.pass_context
@handle_error
def findsharedlinks(ctx, keywords):
    """查找本地保存的分享连接

    注意: 使用这个命令必须将配置文件 ~/.alipcs-py/config.toml 中的 [share] store 设为 true
    """

    _find_shared_links(keywords)


def _find_shared_files(keywords: List[str], share_ids: List[str] = [], verbose: bool = False):
    shared_store = SharedStore()
    shared_files = shared_store.search_shared_files(*keywords, share_ids=share_ids)
    display_shared_files(*shared_files, verbose=verbose)


@app.command()
@click.argument("keywords", nargs=-1, type=str)
@click.option("--share-id", "--si", multiple=True, type=str, help="要搜索的 share id")
@click.option("--verbose", "-v", is_flag=True, help="显示细节")
@click.pass_context
@handle_error
def findsharedfiles(ctx, keywords, share_id, verbose):
    """查找本地保存的分享文件

    注意: 使用这个命令必须将配置文件 ~/.alipcs-py/config.toml 中的 [share] store 设为 true
    """

    _find_shared_files(keywords, share_ids=share_id, verbose=verbose)


@app.command()
@click.argument("keywords", nargs=-1, type=str)
@click.option("--verbose", "-v", is_flag=True, help="显示细节")
@click.pass_context
@handle_error
def findshared(ctx, keywords, verbose):
    """查找本地保存的分享连接和文件

    注意: 使用这个命令必须将配置文件 ~/.alipcs-py/config.toml 中的 [share] store 设为 true
    """

    _find_shared_links(keywords)
    print()
    _find_shared_files(keywords, verbose=verbose)


@app.command()
@click.argument("share_ids", nargs=-1, type=str)
@click.option("--keyword", "-k", multiple=True, type=str, help="要删除文件名的关键字，如果为空则删除share_id下的所有文件")
@click.pass_context
@handle_error
@save_modified_user_data
def deletestoredshared(ctx, share_ids, keyword):
    """删除本地保存的分享连接或文件

    注意: 使用这个命令必须将配置文件 ~/.alipcs-py/config.toml 中的 [share] store 设为 true
    """

    if not share_ids:
        return

    shared_store = SharedStore()

    if not keyword:
        shared_links = []
        for share_id in share_ids:
            sli = shared_store.get_shared_link_info(share_id)
            if sli:
                shared_links.append(sli)
        display_shared_link_infos(*shared_links)
        yes = Prompt.ask(
            "Are you sure to delete these stored shared links and their files?",
            choices=["n", "Yes"],
            default="n",
        )
        if yes == "Yes":
            shared_store.delete_shared_links(*share_ids)
    else:
        shared_files = shared_store.search_shared_files(*keyword, share_ids=share_ids)
        display_shared_files(*shared_files)
        yes = Prompt.ask(
            "Are you sure to delete these stored shared files?",
            choices=["n", "Yes"],
            default="n",
        )
        if yes == "Yes":
            file_ids = [sf.file_id for sf, _ in shared_files]
            shared_store.delete_shared_files(*file_ids)


@app.command()
@click.pass_context
@handle_error
def cleanstore(ctx):
    """清理本地保存的无效分享连接

    注意: 使用这个命令必须将配置文件 ~/.alipcs-py/config.toml 中的 [share] store 设为 true
    """

    app_config = _app_config(ctx)
    if not app_config.share.store:
        print("App configuration `[share] store is false`. So the command does not work")
        return

    api: AliPCSApiMixWithSharedStore = _recent_api(ctx)
    if not api:
        return

    store = api.sharedstore
    if not store:
        return

    shared_links = store.list_shared_links()
    for shared_link in shared_links:
        if not api.is_shared_valid(shared_link.share_id):
            store.delete_shared_links(shared_link.share_id)
            display_invalid_shared_link_infos(shared_link)


# }}}


# Server
# {{{
@app.command()
@click.argument("root_dir", type=str, default="/", required=False)
@click.option("--path", type=str, default="/", help="服务路径，默认为 “/”")
@click.option("--host", "-h", type=str, default="localhost", help="监听 host")
@click.option("--port", "-p", type=int, default=8000, help="监听 port")
@click.option("--workers", "-w", type=int, default=CPU_NUM, help="进程数")
@click.option("--encrypt-password", "--ep", type=str, default=None, help="加密密码，默认使用用户设置的")
@click.option("--username", type=str, default=None, help="HTTP Basic Auth 用户名")
@click.option("--password", type=str, default=None, help="HTTP Basic Auth 密钥")
@click.pass_context
@handle_error
@save_modified_user_data
@multi_user_do
def server(ctx, root_dir, path, host, port, workers, encrypt_password, username, password):
    """开启 HTTP 服务"""

    api = _recent_api(ctx)
    if not api:
        return

    encrypt_password = encrypt_password or _encrypt_password(ctx)

    if username:
        assert password, "Must set password"

    start_server(
        api,
        root_dir=root_dir,
        path=path,
        host=host,
        port=port,
        workers=workers,
        encrypt_password=encrypt_password,
        username=username,
        password=password,
    )


# }}}



PHONE_MODEL_DATABASE = [
    "1501_M02",  # 360 F4
    "1503-M02",  # 360 N4
    "1505-A01",  # 360 N4S
    "303SH",  # 夏普 Aquos Crystal Xx Mini 303SH
    "304SH",  # 夏普 Aquos Crystal Xx SoftBank
    "305SH",  # 夏普 Aquos Crystal Y
    "306SH",  # 夏普 Aquos Crystal 306SH
    "360 Q5 Plus",  # 360 Q5 Plus
    "360 Q5",  # 360 Q5
    "402SH",  # 夏普 Aquos Crystal X
    "502SH",  # 夏普 Aquos Crystal Xx2
    "6607",  # OPPO U3
    "A1001",  # 一加手机1
    "ASUS_A001",  # 华硕 ZenFone 3 Ultra
    "ASUS_A001",  # 华硕 ZenFone 3 Ultra
    "ASUS_Z00ADB",  # 华硕 ZenFone 2
    "ASUS_Z00UDB",  # 华硕 Zenfone Selfie
    "ASUS_Z00XSB",  # 华硕 ZenFone Zoom
    "ASUS_Z012DE",  # 华硕 ZenFone 3
    "ASUS_Z012DE",  # 华硕 ZenFone 3
    "ASUS_Z016D",  # 华硕 ZenFone 3 尊爵
    "ATH-TL00H",  # 华为 荣耀 7i
    "Aster T",  # Vertu Aster T
    "BLN-AL10",  # 华为 荣耀 畅玩6X
    "BND-AL10",  # 荣耀7X
    "BTV-W09",  # 华为 M3
    "CAM-UL00",  # 华为 荣耀 畅玩5A
    "Constellation V",  # Vertu Constellation V
    "D6683",  # 索尼 Xperia Z3 Dual TD
    "DIG-AL00",  # 华为 畅享 6S
    "E2312",  # 索尼 Xperia M4 Aqua
    "E2363 ",  # 索尼 Xperia M4 Aqua Dual
    "E5363",  # 索尼 Xperia C4
    "E5563",  # 索尼 Xperia C5
    "E5663",  # 索尼 Xperia M5
    "E5823",  # 索尼 Xperia Z5 Compact
    "E6533",  # 索尼 Xperia Z3+
    "E6683",  # 索尼 Xperia Z5
    "E6883",  # 索尼 Xperia Z5 Premium
    "EBEN M2",  # 8848 M2
    "EDI-AL10",  # 华为 荣耀 Note 8
    "EVA-AL00",  # 华为 P9
    "F100A",  # 金立 F100
    "F103B",  # 金立 F103B
    "F3116",  # 索尼 Xperia XA
    "F3216",  # 索尼 Xperia XA Ultra
    "F5121 / F5122",  # 索尼 Xperia X
    "F5321",  # 索尼 Xperia X Compact
    "F8132",  # 索尼 Xperia X Performance
    "F8332",  # 索尼 Xperia XZ
    "FRD-AL00",  # 华为 荣耀 8
    "FS8001",  # 夏普 C1
    "FS8002",  # 夏普 A1
    "G0111",  # 格力手机 1
    "G0215",  # 格力手机 2
    "G8142",  # 索尼Xperia XZ Premium G8142
    "G8342",  # 索尼Xperia XZ1
    "GIONEE S9",  # 金立 S9
    "GN5001S",  # 金立 金钢
    "GN5003",  # 金立 大金钢
    "GN8002S",  # 金立 M6 Plus
    "GN8003",  # 金立 M6
    "GN9011",  # 金立 S8
    "GN9012",  # 金立 S6 Pro
    "GRA-A0",  # Coolpad Cool Play 6C
    "H60-L11",  # 华为 荣耀 6
    "HN3-U01",  # 华为 荣耀 3
    "HTC D10w",  # HTC Desire 10 Pro
    "HTC E9pw",  # HTC One E9+
    "HTC M10u",  # HTC 10
    "HTC M8St",  # HTC One M8
    "HTC M9PT",  # HTC One M9+
    "HTC M9e",  # HTC One M9
    "HTC One A9",  # HTC One A9
    "HTC U-1w",  # HTC U Ultra
    "HTC X9u",  # HTC One X9
    "HTC_M10h",  # HTC 10 国际版
    "HUAWEI CAZ-AL00",  # 华为 Nova
    "HUAWEI CRR-UL00",  # 华为 Mate S
    "HUAWEI GRA-UL10",  # 华为 P8
    "HUAWEI MLA-AL10",  # 华为 麦芒 5
    "HUAWEI MT7-AL00",  # 华为 mate 7
    "HUAWEI MT7-TL00",  # 华为 Mate 7
    "HUAWEI NXT-AL10",  # 华为 Mate 8
    "HUAWEI P7-L00",  # 华为 P7
    "HUAWEI RIO-AL00",  # 华为 麦芒 4
    "HUAWEI TAG-AL00",  # 华为 畅享 5S
    "HUAWEI VNS-AL00",  # 华为 G9
    "IUNI N1",  # 艾优尼 N1
    "IUNI i1",  # 艾优尼 i1
    "KFAPWI",  # Amazon Kindle Fire HDX 8.9
    "KFSOWI",  # Amazon Kindle Fire HDX 7
    "KFTHWI",  # Amazon Kindle Fire HD
    "KIW-TL00H",  # 华为 荣耀 畅玩5X
    "KNT-AL10",  # 华为 荣耀 V8
    "L55t",  # 索尼 Xperia Z3
    "L55u",  # 索尼 Xperia Z3
    "LEX626",  # 乐视 乐S3
    "LEX720",  # 乐视 乐Pro3
    "LG-D858",  # LG G3
    "LG-H818",  # LG G4
    "LG-H848",  # LG G5 SE
    "LG-H868",  # LG G5
    "LG-H968",  # LG V10
    "LON-AL00",  # 华为 Mate 9 Pro
    "LON-AL00-PD",  # 华为 Mate 9 Porsche Design
    "LT18i",  # Sony Ericsson Xperia Arc S
    "LT22i",  # Sony Ericsson Xperia P
    "LT26i",  # Sony Ericsson Xperia S
    "LT26ii",  # Sony Ericsson Xperia SL
    "LT26w",  # Sony Ericsson Xperia Acro S
    "Le X520",  # 乐视 乐2
    "Le X620",  # 乐视 乐2Pro
    "Le X820",  # 乐视 乐Max2
    "Lenovo A3580",  # 联想 黄金斗士 A8 畅玩
    "Lenovo A7600-m",  # 联想 黄金斗士 S8
    "Lenovo A938t",  # 联想 黄金斗士 Note8
    "Lenovo K10e70",  # 联想 乐檬K10
    "Lenovo K30-T",  # 联想 乐檬 K3
    "Lenovo K32C36",  # 联想 乐檬3
    "Lenovo K50-t3s",  # 联想 乐檬 K3 Note
    "Lenovo K52-T38",  # 联想 乐檬 K5 Note
    "Lenovo K52e78",  # Lenovo K5 Note
    "Lenovo P2c72",  # 联想 P2
    "Lenovo X3c50",  # 联想 乐檬 X3
    "Lenovo Z90-3",  # 联想 VIBE Shot大拍
    "M040",  # 魅族 MX 2
    "M1 E",  # 魅蓝 E
    "M2-801w",  # 华为 M2
    "M2017",  # 金立 M2017
    "M3",  # EBEN M3
    "M355",  # 魅族 MX 3
    "MHA-AL00",  # 华为 Mate 9
    "MI 4LTE",  # 小米手机4
    "MI 4S",  # 小米手机4S
    "MI 5",  # 小米手机5
    "MI 5s Plus",  # 小米手机5s Plus
    "MI 5s",  # 小米手机5s
    "MI MAX",  # 小米Max
    "MI Note Pro",  # 小米Note顶配版
    "MI PAD 2",  # 小米平板 2
    "MIX",  # 小米MIX
    "MLA-UL00",  # 华为 G9 Plus
    "MP1503",  # 美图 M6
    "MP1512",  # 美图 M6s
    "MT27i",  # Sony Ericsson Xperia Sola
    "MX4 Pro",  # 魅族 MX 4 Pro
    "MX4",  # 魅族 MX 4
    "MX5",  # 魅族 MX 5
    "MX6",  # 魅族 MX 6
    "Meitu V4s",  # 美图 V4s
    "Meizu M3 Max",  # 魅蓝max
    "Meizu U20",  # 魅蓝U20
    "Mi 5",
    "Mi 6",
    "Mi A1",  # MI androidone
    "Mi Note 2",  # 小米Note2
    "MiTV2S-48",  # 小米电视2s
    "Moto G (4)",  # 摩托罗拉 G⁴ Plus
    "N1",  # Nokia N1
    "NCE-AL00",  # 华为 畅享 6
    "NTS-AL00",  # 华为 荣耀 Magic
    "NWI-AL10",  # nova2s
    "NX508J",  # 努比亚 Z9
    "NX511J",  # 努比亚 小牛4 Z9 Mini
    "NX512J",  # 努比亚 大牛 Z9 Max
    "NX513J",  # 努比亚 My 布拉格
    "NX513J",  # 努比亚 布拉格S
    "NX523J",  # 努比亚 Z11 Max
    "NX529J",  # 努比亚 小牛5 Z11 Mini
    "NX531J",  # 努比亚 Z11
    "NX549J",  # 努比亚 小牛6 Z11 MiniS
    "NX563J",  # 努比亚Z17
    "Nexus 4",
    "Nexus 5X",
    "Nexus 6",
    "Nexus 6P",
    "Nexus 7",
    "Nexus 9",
    "Nokia_X",  # Nokia X
    "Nokia_XL_4G",  # Nokia XL
    "ONE A2001",  # 一加手机2
    "ONE E1001",  # 一加手机X
    "ONEPLUS A5010",  # 一加5T
    "OPPO A53",  # OPPO A53
    "OPPO A59M",  # OPPO A59
    "OPPO A59s",  # OPPO A59s
    "OPPO R11",
    "OPPO R7",  # OPPO R7
    "OPPO R7Plus",  # OPPO R7Plus
    "OPPO R7S",  # OPPO R7S
    "OPPO R7sPlus",  # OPPO R7sPlus
    "OPPO R9 Plustm A",  # OPPO R9Plus
    "OPPO R9s Plus",  # OPPO R9s Plus
    "OPPO R9s",
    "OPPO R9s",  # OPPO R9s
    "OPPO R9tm",  # OPPO R9
    "PE-TL10",  # 华为 荣耀 6 Plus
    "PLK-TL01H",  # 华为 荣耀 7
    "Pro 5",  # 魅族 Pro 5
    "Pro 6",  # 魅族 Pro 6
    "Pro 6s",  # 魅族 Pro 6s
    "RM-1010",  # Nokia Lumia 638
    "RM-1018",  # Nokia Lumia 530
    "RM-1087",  # Nokia Lumia 930
    "RM-1090",  # Nokia Lumia 535
    "RM-867",  # Nokia Lumia 920
    "RM-875",  # Nokia Lumia 1020
    "RM-887",  # Nokia Lumia 720
    "RM-892",  # Nokia Lumia 925
    "RM-927",  # Nokia Lumia 929
    "RM-937",  # Nokia Lumia 1520
    "RM-975",  # Nokia Lumia 635
    "RM-977",  # Nokia Lumia 630
    "RM-984",  # Nokia Lumia 830
    "RM-996",  # Nokia Lumia 1320
    "Redmi 3S",  # 红米3s
    "Redmi 4",  # 小米 红米4
    "Redmi 4A",  # 小米 红米4A
    "Redmi Note 2",  # 小米 红米Note2
    "Redmi Note 3",  # 小米 红米Note3
    "Redmi Note 4",  # 小米 红米Note4
    "Redmi Pro",  # 小米 红米Pro
    "S3",  # 佳域S3
    "SCL-TL00H",  # 华为 荣耀 4A
    "SD4930UR",  # Amazon Fire Phone
    "SH-03G",  # 夏普 Aquos Zeta SH-03G
    "SH-04F",  # 夏普 Aquos Zeta SH-04F
    "SHV31",  # 夏普 Aquos Serie Mini SHV31
    "SM-A5100",  # Samsung Galaxy A5
    "SM-A7100",  # Samsung Galaxy A7
    "SM-A8000",  # Samsung Galaxy A8
    "SM-A9000",  # Samsung Galaxy A9
    "SM-A9100",  # Samsung Galaxy A9 高配版
    "SM-C5000",  # Samsung Galaxy C5
    "SM-C5010",  # Samsung Galaxy C5 Pro
    "SM-C7000",  # Samsung Galaxy C7
    "SM-C7010",  # Samsung Galaxy C7 Pro
    "SM-C9000",  # Samsung Galaxy C9 Pro
    "SM-G1600",  # Samsung Galaxy Folder
    "SM-G5500",  # Samsung Galaxy On5
    "SM-G6000",  # Samsung Galaxy On7
    "SM-G7100",  # Samsung Galaxy On7(2016)
    "SM-G7200",  # Samsung Galasy Grand Max
    "SM-G9198",  # Samsung 领世旗舰Ⅲ
    "SM-G9208",  # Samsung Galaxy S6
    "SM-G9250",  # Samsung Galasy S7 Edge
    "SM-G9280",  # Samsung Galaxy S6 Edge+
    "SM-G9300",  # Samsung Galaxy S7
    "SM-G9350",  # Samsung Galaxy S7 Edge
    "SM-G9500",  # Samsung Galaxy S8
    "SM-G9550",  # Samsung Galaxy S8+
    "SM-G9600",  # Samsung Galaxy S9
    "SM-G960F",  # Galaxy S9 Dual SIM
    "SM-G9650",  # Samsung Galaxy S9+
    "SM-G965F",  # Galaxy S9+ Dual SIM
    "SM-J3109",  # Samsung Galaxy J3
    "SM-J3110",  # Samsung Galaxy J3 Pro
    "SM-J327A",  # Samsung Galaxy J3 Emerge
    "SM-J5008",  # Samsung Galaxy J5
    "SM-J7008",  # Samsung Galaxy J7
    "SM-N9108V",  # Samsung Galasy Note4
    "SM-N9200",  # Samsung Galaxy Note5
    "SM-N9300",  # Samsung Galaxy Note 7
    "SM-N935S",  # Samsung Galaxy Note Fan Edition
    "SM-N9500",  # Samsung Galasy Note8
    "SM-W2015",  # Samsung W2015
    "SM-W2016",  # Samsung W2016
    "SM-W2017",  # Samsung W2017
    "SM705",  # 锤子 T1
    "SM801",  # 锤子 T2
    "SM901",  # 锤子 M1
    "SM919",  # 锤子 M1L
    "ST18i",  # Sony Ericsson Xperia Ray
    "ST25i",  # Sony Ericsson Xperia U
    "STV100-1",  # 黑莓Priv
    "Signature Touch",  # Vertu Signature Touch
    "TA-1000",  # Nokia 6
    "TA-1000",  # HMD Nokia 6
    "TA-1041",  # Nokia 7
    "VERTU Ti",  # Vertu Ti
    "VIE-AL10",  # 华为 P9 Plus
    "VIVO X20",
    "VIVO X20A",
    "W909",  # 金立 天鉴 W909
    "X500",  # 乐视 乐1S
    "X608",  # 乐视 乐1
    "X800",  # 乐视 乐1Pro
    "X900",  # 乐视 乐Max
    "XT1085",  # 摩托罗拉 X
    "XT1570",  # 摩托罗拉 X Style
    "XT1581",  # 摩托罗拉 X 极
    "XT1585",  # 摩托罗拉 Droid Turbo 2
    "XT1635",  # 摩托罗拉 Z Play
    "XT1635-02",  # 摩托罗拉 Z Play
    "XT1650",  # 摩托罗拉 Z
    "XT1650-05",  # 摩托罗拉 Z
    "XT1706",  # 摩托罗拉 E³ POWER
    "YD201",  # YotaPhone2
    "YD206",  # YotaPhone2
    "YQ60",  # 锤子 坚果
    "ZTE A2015",  # 中兴 AXON 天机
    "ZTE A2017",  # 中兴 AXON 天机 7
    "ZTE B2015",  # 中兴 AXON 天机 MINI
    "ZTE BV0720",  # 中兴 Blade A2
    "ZTE BV0730",  # 中兴 Blade A2 Plus
    "ZTE C2016",  # 中兴 AXON 天机 MAX
    "ZTE C2017",  # 中兴 AXON 天机 7 MAX
    "ZTE G720C",  # 中兴 星星2号
    "ZUK Z2121",  # ZUK Z2 Pro
    "ZUK Z2131",  # ZUK Z2
    "ZUK Z2151",  # ZUK Edge
    "ZUK Z2155",  # ZUK Edge L
    "m030",  # 魅族mx
    "m1 metal",  # 魅蓝metal
    "m1 note",  # 魅蓝 Note
    "m1",  # 魅蓝
    "m2 note",  # 魅蓝 Note 2
    "m2",  # 魅蓝 2
    "m3 note",  # 魅蓝 Note 3
    "m3",  # 魅蓝 3
    "m3s",  # 魅蓝 3S
    "m9",  # 魅族m9
    "marlin",  # Google Pixel XL
    "sailfish",  # Google Pixel
    "vivo V3Max",  # vivo V3Max
    "vivo X6D",  # vivo X6
    "vivo X6PlusD",  # vivo X6Plus
    "vivo X6S",  # vivo X6S
    "vivo X6SPlus",  # vivo X6SPlus
    "vivo X7",  # vivo X7
    "vivo X7Plus",  # vivo X7Plus
    "vivo X9",  # vivo X9
    "vivo X9Plus",  # vivo X9Plus
    "vivo Xplay5A 金",  # vivo Xplay5
    "vivo Xplay6",  # vivo Xplay6
    "vivo Y66",  # vivo Y66
    "vivo Y67",  # vivo Y67
    "z1221",  # ZUK Z1
]


# 根据key计算出imei
def sum_IMEI(key: str):
    hs = 53202347234687234
    for k in key:
        hs += (hs << 5) + ord(k)
    hs %= int(1e15)
    if hs < int(1e14):
        hs += int(1e14)
    return str(int(hs))


# 根据key, 从 PHONE_MODEL_DATABASE 中取出手机型号
def get_phone_model(key: str):
    if len(PHONE_MODEL_DATABASE) <= 0:
        return "S3"
    hs = 2134
    for k in key:
        hs += (hs << 4) + ord(k)
    hs %= len(PHONE_MODEL_DATABASE)
    return PHONE_MODEL_DATABASE[hs]



from typing import Optional, List, Dict
from enum import Enum
from pathlib import Path
import os
import shutil
import subprocess
import random
import time
from urllib.parse import quote

from alipcs_py.alipcs import AliPCSApi, PcsFile
from alipcs_py.commands.sifter import Sifter, sift
from alipcs_py.commands.download import USER_AGENT
from alipcs_py.commands.errors import CommandError
from alipcs_py.common.file_type import MEDIA_EXTS
from alipcs_py.common.path import join_path

from rich import print


_print = __builtins__["print"]  # type: ignore


def _with_media_ext(path: str) -> bool:
    ext = os.path.splitext(path)[-1].lower()
    if ext in MEDIA_EXTS:
        return True
    else:
        return False


class Player(Enum):
    mpv = "mpv"  # https://mpv.io

    def which(self) -> Optional[str]:
        return shutil.which(self.value)

    def play(
        self,
        url: str,
        quiet: bool = False,
        player_params: List[str] = [],
        out_cmd: bool = False,
        use_local_server: bool = False,
    ):
        global DEFAULT_PLAYER
        if not self.which():
            print(f"[yellow]No player {self.name}[/yellow], using default player: {DEFAULT_PLAYER.name}")
            self = DEFAULT_PLAYER
        if not self.which():
            raise CommandError(f"No player: {self.name}")

        if self == Player.mpv:
            cmd = self._mpv_cmd(
                url,
                quiet=quiet,
                player_params=player_params,
                use_local_server=use_local_server,
            )
        else:
            cmd = self._mpv_cmd(
                url,
                quiet=quiet,
                player_params=player_params,
                use_local_server=use_local_server,
            )

        # Print out command
        if out_cmd:
            _print(" ".join((repr(c) for c in cmd)))
            return

        returncode = self.spawn(cmd)
        if returncode != 0:
            print(f"[italic]{self.value}[/italic] fails. return code: [red]{returncode}[/red]")

    def spawn(self, cmd: List[str], quiet: bool = False):
        child = subprocess.run(
            cmd,
            stdout=subprocess.DEVNULL if quiet else None,
        )
        return child.returncode

    def _mpv_cmd(
        self,
        url: str,
        quiet: bool = False,
        player_params: List[str] = [],
        use_local_server: bool = False,
    ):
        if use_local_server:
            cmd = [self.which(), url, *player_params]
        else:
            cmd = [
                self.which(),
                url,
                "--no-ytdl",
                "--http-header-fields="
                f"User-Agent: {USER_AGENT.replace(',', '')},Referer: https://www.aliyundrive.com/,Connection: Keep-Alive",
                *player_params,
            ]
        if quiet:
            cmd.append("--really-quiet")
        return cmd


DEFAULT_PLAYER = Player.mpv


def play_file(
    api: AliPCSApi,
    pcs_file: PcsFile,
    share_id: str = None,
    player: Player = DEFAULT_PLAYER,
    player_params: List[str] = [],
    quiet: bool = False,
    ignore_ext: bool = False,
    out_cmd: bool = False,
    local_server: str = "",
):
    if not ignore_ext and not _with_media_ext(pcs_file.name):
        return

    print(f"[italic blue]Play[/italic blue]: {pcs_file.path or pcs_file.name}")

    # For typing
    download_url: Optional[str] = None

    use_local_server = bool(local_server)

    if share_id:
        use_local_server = False
        remote_temp_dir = "/__alipcs_py_temp__"
        pcs_temp_dir = api.path(remote_temp_dir) or api.makedir_path(remote_temp_dir)
        pcs_file = api.transfer_shared_files([pcs_file.file_id], pcs_temp_dir.file_id, share_id)[0]
        # download_url = api.shared_file_download_url(pcs_file.file_id, share_id)

        while True:
            pcs_file = api.meta(pcs_file.file_id)[0]
            if pcs_file.download_url:
                break
            time.sleep(2)

    if use_local_server:
        download_url = f"{local_server}/__fileid__/?file_id={pcs_file.file_id}"
        print("url:", download_url)
    else:
        if not pcs_file or pcs_file.is_dir:
            return
        pcs_file.update_download_url(api)
        download_url = pcs_file.download_url

    if download_url:
        player.play(
            download_url,
            quiet=quiet,
            player_params=player_params,
            out_cmd=out_cmd,
            use_local_server=use_local_server,
        )

    if share_id:
        api.remove(pcs_file.file_id)


def play_dir(
    api: AliPCSApi,
    pcs_file: PcsFile,
    share_id: str = None,
    sifters: List[Sifter] = [],
    recursive: bool = False,
    from_index: int = 0,
    player: Player = DEFAULT_PLAYER,
    player_params: List[str] = [],
    quiet: bool = False,
    shuffle: bool = False,
    ignore_ext: bool = False,
    out_cmd: bool = False,
    local_server: str = "",
):
    if pcs_file.path.startswith("/"):
        remotefiles = list(api.list_path(pcs_file.path, share_id=share_id))
    else:
        remotefiles = list(api.list_iter(pcs_file.file_id, share_id=share_id))
    remotefiles = sift(remotefiles, sifters, recursive=recursive)

    if shuffle:
        rg = random.Random(time.time())
        rg.shuffle(remotefiles)

    for rp in remotefiles[from_index:]:
        if rp.is_file:
            play_file(
                api,
                rp,
                share_id=share_id,
                player=player,
                player_params=player_params,
                quiet=quiet,
                ignore_ext=ignore_ext,
                out_cmd=out_cmd,
                local_server=local_server,
            )
        else:  # is_dir
            if recursive:
                play_dir(
                    api,
                    rp,
                    share_id=share_id,
                    sifters=sifters,
                    recursive=recursive,
                    from_index=from_index,
                    player=player,
                    player_params=player_params,
                    quiet=quiet,
                    shuffle=shuffle,
                    ignore_ext=ignore_ext,
                    out_cmd=out_cmd,
                    local_server=local_server,
                )


def play(
    api: AliPCSApi,
    remotepaths: List[str],
    file_ids: List[str],
    share_id: str = None,
    sifters: List[Sifter] = [],
    recursive: bool = False,
    from_index: int = 0,
    player: Player = DEFAULT_PLAYER,
    player_params: List[str] = [],
    quiet: bool = False,
    shuffle: bool = False,
    ignore_ext: bool = False,
    out_cmd: bool = False,
    local_server: str = "",
):
    """Play media file in `remotepaths`

    Args:
        `from_index` (int): The start index of playing entries from EACH remote directory
    """

    if shuffle:
        rg = random.Random(time.time())
        rg.shuffle(remotepaths)

    for rp in remotepaths:
        rpf = api.path(rp, share_id=share_id)
        if not rpf:
            print(f"[yellow]WARNING[/yellow]: `{rp}` does not exist.")
            continue

        if rpf.is_file:
            play_file(
                api,
                rpf,
                share_id=share_id,
                player=player,
                player_params=player_params,
                quiet=quiet,
                ignore_ext=ignore_ext,
                out_cmd=out_cmd,
                local_server=local_server,
            )
        else:
            play_dir(
                api,
                rpf,
                share_id=share_id,
                sifters=sifters,
                recursive=recursive,
                from_index=from_index,
                player=player,
                player_params=player_params,
                quiet=quiet,
                shuffle=shuffle,
                ignore_ext=ignore_ext,
                out_cmd=out_cmd,
                local_server=local_server,
            )

    for file_id in file_ids:
        rpf = api.meta(file_id, share_id=share_id)[0]
        if not rpf:
            print(f"[yellow]WARNING[/yellow]: file_id `{file_id}` does not exist.")
            continue

        if rpf.is_file:
            play_file(
                api,
                rpf,
                share_id=share_id,
                player=player,
                player_params=player_params,
                quiet=quiet,
                ignore_ext=ignore_ext,
                out_cmd=out_cmd,
                local_server=local_server,
            )
        else:
            play_dir(
                api,
                rpf,
                share_id=share_id,
                sifters=sifters,
                recursive=recursive,
                from_index=from_index,
                player=player,
                player_params=player_params,
                quiet=quiet,
                shuffle=shuffle,
                ignore_ext=ignore_ext,
                out_cmd=out_cmd,
                local_server=local_server,
            )



from typing import Optional, List, Tuple, Dict, Union, DefaultDict, Iterable, Iterator, Callable, IO
from threading import Lock
from collections import defaultdict
from copy import deepcopy
from functools import partial
import logging

from alipcs_py.alipcs.errors import AliPCSError
from alipcs_py.common.io import RangeRequestIO, DEFAULT_MAX_CHUNK_SIZE
from alipcs_py.alipcs.pcs import AliPCS, AliOpenPCS, CheckNameMode
from alipcs_py.alipcs.inner import PcsFile, PcsPreparedFile, PcsSharedLink, PcsSharedLinkInfo, PcsUser, PcsDownloadUrl
from alipcs_py.common.path import split_posix_path, join_path, posix_path_dirname

from requests_toolbelt import MultipartEncoderMonitor

logger = logging.getLogger(__name__)


_ALI_PCS_API_LOCK = Lock()


class AliPCSApi:
    """Aliyundrive PCS Api

    This is the wrapper of `AliPCS`. It parses the content of response of raw
    AliPCS requests to some inner data structions.
    """

    def __init__(
        self,
        refresh_token: str,
        access_token: str = "",
        token_type: str = "",
        expire_time: int = 0,
        user_id: str = "",
        user_name: str = "",
        nick_name: str = "",
        device_id: str = "",
        default_drive_id: str = "",
        role: str = "",
        status: str = "",
    ):
        self._alipcs = AliPCS(
            refresh_token,
            access_token=access_token,
            token_type=token_type,
            expire_time=expire_time,
            user_id=user_id,
            user_name=user_name,
            nick_name=nick_name,
            device_id=device_id,
            default_drive_id=default_drive_id,
            role=role,
            status=status,
        )

        # The path tree is for user's own files
        self._path_tree = PathTree(self)

        # Map `share_id` to a `PathTree` for that shared link
        self._shared_path_tree: DefaultDict[str, PathTree] = defaultdict(partial(PathTree, self))

    @property
    def refresh_token(self) -> str:
        return self._alipcs.refresh_token

    @property
    def access_token(self) -> str:
        return self._alipcs.access_token

    @property
    def token_type(self) -> str:
        return self._alipcs.token_type

    @property
    def expire_time(self) -> int:
        return self._alipcs.expire_time

    @property
    def user_id(self) -> str:
        return self._alipcs.user_id

    @property
    def user_name(self) -> str:
        return self._alipcs.user_name

    @property
    def nick_name(self) -> str:
        return self._alipcs.nick_name

    @property
    def device_id(self) -> str:
        return self._alipcs.device_id

    @property
    def default_drive_id(self) -> str:
        return self._alipcs.default_drive_id

    @property
    def role(self) -> str:
        return self._alipcs.role

    @property
    def status(self) -> str:
        return self._alipcs.status

    def meta(self, *file_ids: str, share_id: str = None) -> List[PcsFile]:
        """Meta data of `remotepaths`"""

        pcs_files = [PcsFile.root() if fid == "root" else None for fid in file_ids]
        fids = [fid for fid in file_ids if fid != "root"]

        if fids:
            info = self._alipcs.meta(*fids, share_id=share_id)
            pfs = [PcsFile.from_(v.get("body")) for v in info["responses"]]
            j = 0
            for i in range(len(pcs_files)):
                if pcs_files[i] is None:
                    pcs_files[i] = pfs[j]
                    j += 1

        return [pf for pf in pcs_files if pf is not None]

    def exists(self, file_id: str) -> bool:
        """Check whether `remotepath` exists"""

        return self._alipcs.exists(file_id)

    def is_file(self, file_id: str) -> bool:
        """Check whether `remotepath` is a file"""

        return self._alipcs.is_file(file_id)

    def is_dir(self, file_id: str) -> bool:
        """Check whether `remotepath` is a directory"""

        return self._alipcs.is_dir(file_id)

    def list(
        self,
        file_id: str,
        share_id: str = None,
        desc: bool = False,
        name: bool = False,
        time: bool = False,
        size: bool = False,
        all: bool = False,
        limit: int = 200,
        url_expire_sec: int = 14400,
        next_marker: str = "",
    ) -> Tuple[List[PcsFile], str]:
        """List the directory's contents

        List files and directories in the given directory (which has the `file_id`).
        The return items size is limited by the `limit` parameter. If you want to list
        more, using the returned `next_marker` parameter for next `list` call.
        """

        info = self._alipcs.list(
            file_id=file_id,
            share_id=share_id,
            desc=desc,
            name=name,
            time=time,
            size=size,
            all=all,
            limit=limit,
            url_expire_sec=url_expire_sec,
            next_marker=next_marker,
        )
        next_marker = info["next_marker"]
        return [PcsFile.from_(v) for v in info.get("items", [])], next_marker

    def list_iter(
        self,
        file_id: str,
        share_id: str = None,
        desc: bool = False,
        name: bool = False,
        time: bool = False,
        size: bool = False,
        all: bool = False,
        limit: int = 200,
        url_expire_sec: int = 14400,
        recursive: bool = False,
        include_dir: bool = True,
    ) -> Iterator[PcsFile]:
        """Iterate the directory by its `file_id`

        Iterate all files and directories at the directory (which has the `file_id`).
        """

        next_marker = ""
        pcs_file = self.meta(file_id, share_id=share_id)[0]
        dirname = pcs_file.name
        while True:
            pcs_files, next_marker = self.list(
                file_id,
                share_id=share_id,
                desc=desc,
                name=name,
                time=time,
                size=size,
                all=all,
                limit=limit,
                url_expire_sec=url_expire_sec,
                next_marker=next_marker,
            )
            for pf in pcs_files:
                pf.path = join_path(dirname, pf.name)

                if pf.is_dir:
                    if include_dir:
                        yield pf
                    if recursive:
                        for sub_pf in self.list_iter(
                            pf.file_id,
                            share_id=share_id,
                            desc=desc,
                            name=name,
                            time=time,
                            size=size,
                            all=all,
                            limit=limit,
                            url_expire_sec=url_expire_sec,
                            recursive=recursive,
                            include_dir=include_dir,
                        ):
                            sub_pf.path = join_path(dirname, sub_pf.path)
                            yield sub_pf
                else:
                    yield pf

            if not next_marker:
                return

    def path(self, remotepath: str, share_id: str = None) -> Optional[PcsFile]:
        """Get the pcs file's info by the given absolute `remotepath`"""

        if share_id:
            return self._shared_path_tree[share_id].search(remotepath=remotepath, share_id=share_id)
        else:
            return self._path_tree.search(remotepath=remotepath)

    def paths(self, *remotepaths: str, share_id: str = None) -> List[Optional[PcsFile]]:
        """Get the pcs files' info by the given absolute `remotepaths`"""

        return [self.path(rp, share_id=share_id) for rp in remotepaths]

    def list_path_iter(
        self,
        remotepath: str,
        file_id: str = None,
        share_id: str = None,
        desc: bool = False,
        name: bool = False,
        time: bool = False,
        size: bool = False,
        all: bool = False,
        limit: int = 200,
        url_expire_sec: int = 14400,
        recursive: bool = False,
        include_dir: bool = True,
    ) -> Iterator[PcsFile]:
        """Iterate the `remotepath`"""

        if not file_id:
            pf = self.path(remotepath, share_id=share_id)
            if not pf:
                return
            file_id = pf.file_id

        dirname = posix_path_dirname(remotepath)

        for p in self.list_iter(
            file_id,
            share_id=share_id,
            desc=desc,
            name=name,
            time=time,
            size=size,
            all=all,
            limit=limit,
            url_expire_sec=url_expire_sec,
            recursive=recursive,
            include_dir=include_dir,
        ):
            p.path = join_path(dirname, p.path)
            yield p

    def list_path(
        self,
        remotepath: str,
        file_id: str = None,
        share_id: str = None,
        desc: bool = False,
        name: bool = False,
        time: bool = False,
        size: bool = False,
        all: bool = False,
        limit: int = 200,
        url_expire_sec: int = 14400,
    ) -> List[PcsFile]:
        return list(
            self.list_path_iter(
                remotepath,
                file_id=file_id,
                share_id=share_id,
                desc=desc,
                name=name,
                time=time,
                size=size,
                all=all,
                limit=limit,
                url_expire_sec=url_expire_sec,
            )
        )

    def create_file(
        self,
        filename: str,
        dir_id: str,
        size: int,
        pre_hash: str = "",
        content_hash: str = "",
        part_number: int = 1,
        proof_code: str = "",
        check_name_mode: CheckNameMode = "auto_rename",
    ) -> PcsPreparedFile:
        """Create a prepared file for uploading

        filename (str):
            The name of file.
        dir_id (str):
            The directory id where the file is at.
        size (int):
            the length of total content.
        pre_hash (str):
            The sha1 of the IO first 1k bytes.
        content_hash (str):
            the sha1 of total content.
        part_number (int):
            The number of one file's chunks to upload.
            The server will returns the number of urls to prepare to upload the file's chunks.
            `WARNNING`: this value MUST be set by caller.
        check_name_mode(str):
            'overwrite' (直接覆盖，以后多版本有用)
            'auto_rename' (自动换一个随机名称)
            'refuse' (不会创建，告诉你已经存在)
            'ignore' (会创建重名的)
        """

        info = self._alipcs.create_file(
            filename,
            dir_id,
            size,
            pre_hash=pre_hash,
            content_hash=content_hash,
            part_number=part_number,
            proof_code=proof_code,
            check_name_mode=check_name_mode,
        )
        return PcsPreparedFile.from_(info)

    def prepare_file(
        self,
        filename: str,
        dir_id: str,
        size: int,
        pre_hash: str,
        part_number: int = 1,
        check_name_mode: CheckNameMode = "auto_rename",
    ) -> PcsPreparedFile:
        """Create a prepared file with `pre_hash` for uploading"""

        return self.create_file(
            filename, dir_id, size, pre_hash=pre_hash, part_number=part_number, check_name_mode=check_name_mode
        )

    def rapid_upload_file(
        self,
        filename: str,
        dir_id: str,
        size: int,
        content_hash: str,
        proof_code: str,
        check_name_mode: CheckNameMode = "auto_rename",
    ) -> PcsPreparedFile:
        return self.create_file(
            filename, dir_id, size, content_hash=content_hash, proof_code=proof_code, check_name_mode=check_name_mode
        )

    def upload_slice(self, io: IO, url: str, callback: Callable[[MultipartEncoderMonitor], None] = None) -> None:
        """Upload an io as a slice

        callable: the callback for monitoring uploading progress
        """

        self._alipcs.upload_slice(io, url, callback=callback)

    def upload_complete(self, file_id: str, upload_id: str) -> PcsFile:
        """Tell server that all slices have been uploaded

        After uploading all slices using `AliPCSApi.upload_slice`, we need to
        tell server that all slice have been uploaded. Then, the server will
        return the final file info back.
        """

        info = self._alipcs.upload_complete(file_id, upload_id)
        return PcsFile.from_(info)

    def search(
        self,
        keyword: str,
        desc: bool = False,
        name: bool = False,
        time: bool = False,
        size: bool = False,
        all: bool = False,
        limit: int = 100,
        next_marker: str = "",
    ) -> Tuple[List[PcsFile], str]:
        """Search with `keyword`"""

        info = self._alipcs.search(
            keyword, desc=desc, name=name, time=time, size=size, all=all, limit=limit, next_marker=next_marker
        )
        next_marker = info["next_marker"]
        return [PcsFile.from_(v) for v in info.get("items", [])], next_marker

    def search_all(
        self,
        keyword: str,
        desc: bool = False,
        name: bool = False,
        time: bool = False,
        size: bool = False,
        all: bool = False,
        limit: int = 100,
    ) -> List[PcsFile]:
        """Search with `keyword`"""

        next_marker = ""
        pcs_files = []
        while True:
            pfs, next_marker = self.search(
                keyword=keyword,
                desc=desc,
                name=name,
                time=time,
                size=size,
                all=all,
                limit=limit,
                next_marker=next_marker,
            )
            pcs_files.extend(pfs)

            if not next_marker:
                break

        return pcs_files

    def makedir(self, dir_id: str, name: str) -> PcsFile:
        info = self._alipcs.makedir(dir_id, name)
        return PcsFile.from_(info)

    def makedir_path(self, remotedir: str) -> PcsFile:
        # Use lock to ignore make a directory twice
        with _ALI_PCS_API_LOCK:
            parts = split_posix_path(remotedir)
            parent = PcsFile.root()
            for i, part in enumerate(parts):
                if part == "/":  # ignore root
                    continue

                # Find the sub directory which has the name of `part`
                pf = PcsFile.root()
                for pf in self.list_iter(parent.file_id):
                    if pf.name == part:
                        break

                now_dir = join_path(*parts[: i + 1])
                if pf.name == part:
                    assert pf.is_dir, f"{now_dir} is a file"
                else:
                    pf = self.makedir(parent.file_id, part)

                pf.path = now_dir
                parent = pf
            return parent

    def move(self, *file_ids: str) -> List[bool]:
        """Move `file_ids[:-1]` to `file_ids[-1]`"""

        info = self._alipcs.move(*file_ids)

        # Remove nodes from self._path_tree
        for file_id in file_ids[:-1]:
            self._path_tree.pop_by_file_id(file_id)

        return ["code" not in v["body"] for v in info["responses"]]

    def rename(self, file_id: str, name: str) -> PcsFile:
        info = self._alipcs.rename(file_id, name)

        # Remove node from self._path_tree
        self._path_tree.pop_by_file_id(file_id)

        return PcsFile.from_(info)

    def copy(self, *file_ids: str) -> List[PcsFile]:
        """Copy `remotepaths[:-1]` to `remotepaths[-1]`"""

        info = self._alipcs.copy(*file_ids)
        return [PcsFile.from_(v["body"]) for v in info["responses"]]

    def remove(self, *file_ids: str) -> List[bool]:
        """Remove all `remotepaths`"""

        info = self._alipcs.remove(*file_ids)

        # Remove nodes from self._path_tree
        for file_id in file_ids[:-1]:
            self._path_tree.pop_by_file_id(file_id)

        return ["code" not in v for v in info["responses"]]

    def share(self, *file_ids: str, password: str = "", period: int = 0, description: str = "") -> PcsSharedLink:
        """Share `remotepaths` to public with a optional password

        Args:
            period (int): The days for expiring. `0` means no expiring
        """

        info = self._alipcs.share(*file_ids, password=password, period=period, description=description)
        return PcsSharedLink.from_(info)

    def is_shared_valid(self, share_id: str) -> bool:
        try:
            self.shared_info(share_id)
            return True
        except AliPCSError as err:
            if err.error_code in ("ShareLink.Forbidden", "ShareLink.Cancelled", "ShareLink.Expired"):
                return False
            raise

    def list_shared(self, next_marker: str = "") -> Tuple[List[PcsSharedLink], str]:
        """List shared link on a page"""

        info = self._alipcs.list_shared(next_marker=next_marker)
        return [PcsSharedLink.from_(v) for v in info["items"]], info["next_marker"]

    def list_shared_all(self) -> List[PcsSharedLink]:
        """List all shared links"""

        next_marker = ""
        pcs_shared_links = []
        while True:
            pslinks, next_marker = self.list_shared(next_marker=next_marker)
            pcs_shared_links.extend(pslinks)
            if not next_marker:
                return pcs_shared_links

    def cancel_shared(self, *share_ids: str) -> List[bool]:
        """Cancel shared links with their `share_ids`"""

        info = self._alipcs.cancel_shared(*share_ids)
        return ["code" not in v for v in info["responses"]]

    def get_share_token(self, share_id: str, share_password: str = "") -> str:
        """Get share token"""

        info = self._alipcs.get_share_token(share_id, share_password=share_password)
        return info["share_token"]

    def shared_info(self, share_id: str) -> PcsSharedLinkInfo:
        """Get shared link info by anyone"""

        while True:
            try:
                info = self._alipcs.shared_info(share_id)
                return PcsSharedLinkInfo.from_(info)
            except AliPCSError as err:
                # XXX: What is the error?
                if err.error_code == "ParamFlowException":
                    logger.debug("AliPCSApi.shared_info gets error: `ParamFlowException`")
                    continue
                else:
                    raise

    # list_shared_files is an alias of list
    list_shared_files = list

    def transfer_shared_files(
        self, shared_file_ids: List[str], dest_id: str, share_id: str, auto_rename: bool = False
    ) -> List[PcsFile]:
        """Save the `shared_file_ids` to `dest_id`"""

        info = self._alipcs.transfer_shared_files(shared_file_ids, dest_id, share_id, auto_rename=auto_rename)
        return [PcsFile.from_(v["body"]) for v in info.get("responses", [])]

    def shared_file_download_url(self, shared_file_id: str, share_id: str, expire_duration: int = 10 * 60) -> str:
        """Get shared file download link"""

        return self._alipcs.shared_file_download_url(shared_file_id, share_id, expire_duration=expire_duration)

    def user_info(self) -> PcsUser:
        """User's information"""

        info = self._alipcs.user_info()

        user_info = PcsUser.from_(info)
        user_info.web_refresh_token = self.refresh_token
        user_info.web_access_token = self.access_token
        user_info.web_token_type = self.token_type
        user_info.web_expire_time = self.expire_time
        user_info.device_id = self.device_id
        return user_info

    def download_link(self, file_id: str) -> Optional[PcsDownloadUrl]:
        """Get the download link of the `file_id`"""

        info = self._alipcs.download_link(file_id)
        return PcsDownloadUrl.from_(info)

    def file_stream(
        self,
        file_id: str,
        max_chunk_size: int = DEFAULT_MAX_CHUNK_SIZE,
        callback: Callable[..., None] = None,
        encrypt_password: bytes = b"",
    ) -> Optional[RangeRequestIO]:
        """File stream as a normal io"""

        return self._alipcs.file_stream(
            file_id, max_chunk_size=max_chunk_size, callback=callback, encrypt_password=encrypt_password
        )

    def shared_file_stream(
        self,
        shared_file_id: str,
        share_id: str,
        expire_duration: int = 10 * 60,
        max_chunk_size: int = DEFAULT_MAX_CHUNK_SIZE,
        callback: Callable[..., None] = None,
        encrypt_password: bytes = b"",
    ) -> Optional[RangeRequestIO]:
        """Shared file stream as a normal io"""

        return self._alipcs.shared_file_stream(
            shared_file_id,
            share_id,
            expire_duration=expire_duration,
            max_chunk_size=max_chunk_size,
            callback=callback,
            encrypt_password=encrypt_password,
        )


class AliOpenPCSApi:
    """Aliyundrive Open PCS Api

    This is the wrapper of `AliPCS`. It parses the content of response of raw
    AliPCS requests to some inner data structions.
    """

    def __init__(
        self,
        refresh_token: str,
        access_token: str = "",
        expire_time: int = 0,
        client_id: str = "",
        client_secret: str = "",
        client_server: str = "",
        user_id: str = "",
        user_name: str = "",
        nick_name: str = "",
        token_type: str = "",
        default_drive_id: str = "",
        role: str = "",
        status: str = "",
    ):
        self._aliopenpcs = AliOpenPCS(
            refresh_token,
            access_token=access_token,
            expire_time=expire_time,
            client_id=client_id,
            client_secret=client_secret,
            client_server=client_server,
            user_id=user_id,
            user_name=user_name,
            nick_name=nick_name,
            token_type=token_type,
            default_drive_id=default_drive_id,
            role=role,
            status=status,
        )

        # The path tree is for user's own files
        self._path_tree = PathTree(self)

        # Map `share_id` to a `PathTree` for that shared link
        self._shared_path_tree: DefaultDict[str, PathTree] = defaultdict(partial(PathTree, self))

    @property
    def refresh_token(self) -> str:
        return self._aliopenpcs.refresh_token

    @property
    def access_token(self) -> str:
        return self._aliopenpcs.access_token

    @property
    def token_type(self) -> str:
        return self._aliopenpcs.token_type

    @property
    def client_id(self) -> str:
        return self._aliopenpcs.client_id

    @property
    def client_secret(self) -> str:
        return self._aliopenpcs.client_secret

    @property
    def client_server(self) -> str:
        return self._aliopenpcs.client_server

    @property
    def user_id(self) -> str:
        return self._aliopenpcs.user_id

    @property
    def user_name(self) -> str:
        return self._aliopenpcs.user_name

    @property
    def nick_name(self) -> str:
        return self._aliopenpcs.nick_name

    @property
    def expire_time(self) -> int:
        return self._aliopenpcs.expire_time

    @property
    def default_drive_id(self) -> str:
        return self._aliopenpcs.default_drive_id

    @property
    def role(self) -> str:
        return self._aliopenpcs.role

    @property
    def status(self) -> str:
        return self._aliopenpcs.status

    def meta(self, *file_ids: str, share_id: str = None) -> List[PcsFile]:
        """Meta data of `remotepaths`"""

        pcs_files = [PcsFile.root() if fid == "root" else None for fid in file_ids]
        fids = [fid for fid in file_ids if fid != "root"]

        if fids:
            info = self._aliopenpcs.meta(*fids, share_id=share_id)
            pfs = [PcsFile.from_(v.get("body")) for v in info["responses"]]
            j = 0
            for i in range(len(pcs_files)):
                if pcs_files[i] is None:
                    pcs_files[i] = pfs[j]
                    j += 1

        return [pf for pf in pcs_files if pf is not None]

    def exists(self, file_id: str) -> bool:
        """Check whether `remotepath` exists"""

        return self._aliopenpcs.exists(file_id)

    def is_file(self, file_id: str) -> bool:
        """Check whether `remotepath` is a file"""

        return self._aliopenpcs.is_file(file_id)

    def is_dir(self, file_id: str) -> bool:
        """Check whether `remotepath` is a directory"""

        return self._aliopenpcs.is_dir(file_id)

    def list(
        self,
        file_id: str,
        share_id: str = None,
        desc: bool = False,
        name: bool = False,
        time: bool = False,
        size: bool = False,
        all: bool = False,
        limit: int = 200,
        url_expire_sec: int = 14400,
        next_marker: str = "",
    ) -> Tuple[List[PcsFile], str]:
        """List the directory's contents

        List files and directories in the given directory (which has the `file_id`).
        The return items size is limited by the `limit` parameter. If you want to list
        more, using the returned `next_marker` parameter for next `list` call.
        """

        info = self._aliopenpcs.list(
            file_id=file_id,
            share_id=share_id,
            desc=desc,
            name=name,
            time=time,
            size=size,
            all=all,
            limit=limit,
            url_expire_sec=url_expire_sec,
            next_marker=next_marker,
        )
        next_marker = info["next_marker"]
        return [PcsFile.from_(v) for v in info.get("items", [])], next_marker

    def list_iter(
        self,
        file_id: str,
        share_id: str = None,
        desc: bool = False,
        name: bool = False,
        time: bool = False,
        size: bool = False,
        all: bool = False,
        limit: int = 200,
        url_expire_sec: int = 14400,
        recursive: bool = False,
        include_dir: bool = True,
    ) -> Iterator[PcsFile]:
        """Iterate the directory by its `file_id`

        Iterate all files and directories at the directory (which has the `file_id`).
        """

        next_marker = ""
        pcs_file = self.meta(file_id, share_id=share_id)[0]
        dirname = pcs_file.name
        while True:
            pcs_files, next_marker = self.list(
                file_id,
                share_id=share_id,
                desc=desc,
                name=name,
                time=time,
                size=size,
                all=all,
                limit=limit,
                url_expire_sec=url_expire_sec,
                next_marker=next_marker,
            )
            for pf in pcs_files:
                pf.path = join_path(dirname, pf.name)

                if pf.is_dir:
                    if include_dir:
                        yield pf
                    if recursive:
                        for sub_pf in self.list_iter(
                            pf.file_id,
                            share_id=share_id,
                            desc=desc,
                            name=name,
                            time=time,
                            size=size,
                            all=all,
                            limit=limit,
                            url_expire_sec=url_expire_sec,
                            recursive=recursive,
                            include_dir=include_dir,
                        ):
                            sub_pf.path = join_path(dirname, sub_pf.path)
                            yield sub_pf
                else:
                    yield pf

            if not next_marker:
                return

    def path(self, remotepath: str, share_id: str = None) -> Optional[PcsFile]:
        """Get the pcs file's info by the given absolute `remotepath`"""

        if share_id:
            return self._shared_path_tree[share_id].search(remotepath=remotepath, share_id=share_id)
        else:
            return self._path_tree.search(remotepath=remotepath)

    def paths(self, *remotepaths: str, share_id: str = None) -> List[Optional[PcsFile]]:
        """Get the pcs files' info by the given absolute `remotepaths`"""

        return [self.path(rp, share_id=share_id) for rp in remotepaths]

    def list_path_iter(
        self,
        remotepath: str,
        file_id: str = None,
        share_id: str = None,
        desc: bool = False,
        name: bool = False,
        time: bool = False,
        size: bool = False,
        all: bool = False,
        limit: int = 200,
        url_expire_sec: int = 14400,
        recursive: bool = False,
        include_dir: bool = True,
    ) -> Iterator[PcsFile]:
        """Iterate the `remotepath`"""

        if not file_id:
            pf = self.path(remotepath, share_id=share_id)
            if not pf:
                return
            file_id = pf.file_id

        dirname = posix_path_dirname(remotepath)

        for p in self.list_iter(
            file_id,
            share_id=share_id,
            desc=desc,
            name=name,
            time=time,
            size=size,
            all=all,
            limit=limit,
            url_expire_sec=url_expire_sec,
            recursive=recursive,
            include_dir=include_dir,
        ):
            p.path = join_path(dirname, p.path)
            yield p

    def list_path(
        self,
        remotepath: str,
        file_id: str = None,
        share_id: str = None,
        desc: bool = False,
        name: bool = False,
        time: bool = False,
        size: bool = False,
        all: bool = False,
        limit: int = 200,
        url_expire_sec: int = 14400,
    ) -> List[PcsFile]:
        return list(
            self.list_path_iter(
                remotepath,
                file_id=file_id,
                share_id=share_id,
                desc=desc,
                name=name,
                time=time,
                size=size,
                all=all,
                limit=limit,
                url_expire_sec=url_expire_sec,
            )
        )

    def download_link(self, file_id: str) -> Optional[PcsDownloadUrl]:
        """Get the download link of the `file_id`"""

        info = self._aliopenpcs.download_link(file_id)
        return PcsDownloadUrl.from_(info)

    def file_stream(
        self,
        file_id: str,
        max_chunk_size: int = DEFAULT_MAX_CHUNK_SIZE,
        callback: Callable[..., None] = None,
        encrypt_password: bytes = b"",
    ) -> Optional[RangeRequestIO]:
        """File stream as a normal io"""

        return self._aliopenpcs.file_stream(
            file_id, max_chunk_size=max_chunk_size, callback=callback, encrypt_password=encrypt_password
        )


class AliPCSApiMix(AliPCSApi):
    def __init__(
        self,
        web_refresh_token: str,
        web_access_token: str = "",
        web_token_type: str = "",
        web_expire_time: int = 0,
        openapi_refresh_token: str = "",
        openapi_access_token: str = "",
        openapi_token_type: str = "",
        openapi_expire_time: int = 0,
        client_id: str = "",
        client_secret: str = "",
        client_server: str = "",
        user_id: str = "",
        user_name: str = "",
        nick_name: str = "",
        device_id: str = "",
        default_drive_id: str = "",
        role: str = "",
        status: str = "",
    ):
        super().__init__(
            refresh_token=web_refresh_token,
            access_token=web_access_token,
            token_type=web_token_type,
            expire_time=web_expire_time,
            user_id=user_id,
            user_name=user_name,
            nick_name=nick_name,
            device_id=device_id,
            default_drive_id=default_drive_id,
            role=role,
            status=status,
        )

        self._aliopenpcsapi: Optional[AliOpenPCSApi] = None
        if openapi_refresh_token and ((client_id and client_secret) or client_server):
            self._aliopenpcsapi = AliOpenPCSApi(
                refresh_token=openapi_refresh_token,
                access_token=openapi_access_token,
                token_type=openapi_token_type,
                expire_time=openapi_expire_time,
                client_id=client_id,
                client_secret=client_secret,
                client_server=client_server,
                user_id=user_id,
                user_name=user_name,
                nick_name=nick_name,
                default_drive_id=default_drive_id,
                role=role,
                status=status,
            )

    def download_link(self, file_id: str) -> Optional[PcsDownloadUrl]:
        """Get the download link of the `file_id`"""

        if self._aliopenpcsapi:
            return self._aliopenpcsapi.download_link(file_id)
        else:
            return super().download_link(file_id)

    def file_stream(
        self,
        file_id: str,
        max_chunk_size: int = DEFAULT_MAX_CHUNK_SIZE,
        callback: Callable[..., None] = None,
        encrypt_password: bytes = b"",
    ) -> Optional[RangeRequestIO]:
        """File stream as a normal io"""

        if self._aliopenpcsapi:
            return self._aliopenpcsapi.file_stream(
                file_id, max_chunk_size=max_chunk_size, callback=callback, encrypt_password=encrypt_password
            )
        else:
            return super().file_stream(
                file_id, max_chunk_size=max_chunk_size, callback=callback, encrypt_password=encrypt_password
            )

    def user_info(self) -> PcsUser:
        pcs_user = super().user_info()

        if self._aliopenpcsapi:
            pcs_user.openapi_refresh_token = self._aliopenpcsapi.refresh_token
            pcs_user.openapi_access_token = self._aliopenpcsapi.access_token
            pcs_user.openapi_token_type = self._aliopenpcsapi.token_type
            pcs_user.openapi_expire_time = self._aliopenpcsapi.expire_time

            pcs_user.client_id = self._aliopenpcsapi.client_id
            pcs_user.client_secret = self._aliopenpcsapi.client_secret
            pcs_user.client_server = self._aliopenpcsapi.client_server

        return pcs_user


class _Node:
    def __init__(self, file_id: str, pcs_file: PcsFile):
        self.file_id = file_id
        self.pcs_file = pcs_file

        # Map path basename to node
        self.sub_nodes: Dict[str, _Node] = {}

    def _str(self, level: int) -> str:
        s = f"<_Node {self.pcs_file.name}>\n"
        for n in self.sub_nodes.values():
            s = s + " " * ((level + 1) * 2) + n._str(level + 1)
            n._str(level + 1)
        return s

    def __str__(self) -> str:
        return self._str(0)

    def add(self, node: "_Node") -> None:
        """Add a sub node"""

        self.sub_nodes[node.pcs_file.name] = node

    def pop(self, key: str) -> Optional["_Node"]:
        """Pop a sub node which has basename as `key`"""

        try:
            return self.sub_nodes.pop(key)
        except KeyError:
            return None


_PATH_TREE_LOCK = Lock()


class PathTree:
    """Path Tree

    Invite a file or directory by its path is not supported by aliyundrive.
    This class is aim to support to invite a file or directory by its path.
    """

    def __init__(self, api: Union[AliPCSApi, AliOpenPCSApi]):
        self._api = api
        self.root = _Node("root", PcsFile.root())
        self._file_id_to_node: Dict[str, _Node] = {}
        self._file_id_to_node["root"] = self.root

    def __str__(self) -> str:
        return f"<PathTree {self.root} >"

    def _add(self, node: _Node):
        """Add a node to self._file_id_to_node"""

        self._file_id_to_node[node.file_id] = node

    def _pop(self, file_id: str) -> Optional[_Node]:
        """Pop a node from self._file_id_to_node"""

        try:
            return self._file_id_to_node.pop(file_id)
        except KeyError:
            return None

    def search(self, remotepath: str = "", topdown: Iterable[str] = [], share_id: str = None) -> Optional[PcsFile]:
        """Search the PcsFile which has remote path as `remotepath`
        or has the tree path `topdown`
        """

        if not topdown:
            assert remotepath.startswith("/")
            topdown = split_posix_path(remotepath)

        node = self._dfs(list(topdown), self.root, share_id=share_id)
        if node:
            return deepcopy(node.pcs_file)
        else:
            return None

    def _dfs(self, topdown: List[str], root: _Node, pull: bool = True, share_id: str = None) -> Optional[_Node]:
        """Search a node with the path `topdown` using depth first search"""

        if not topdown:
            return root

        next_key = topdown[0]
        if next_key == "/":
            return self._dfs(topdown[1:], self.root, share_id=share_id)

        root_pcs_file = root.pcs_file

        # Add lock
        with _PATH_TREE_LOCK:
            if pull and next_key not in root.sub_nodes:
                for pf in self._api.list_iter(root_pcs_file.file_id, share_id=share_id):
                    if pf.name not in root.sub_nodes:
                        pf.path = join_path(root_pcs_file.path, pf.name)
                        new_node = _Node(pf.file_id, pf)
                        root.add(new_node)
                        self._add(new_node)
                    if pf.name == next_key:
                        break

            if next_key not in root.sub_nodes:
                return None

        return self._dfs(topdown[1:], root.sub_nodes[next_key], share_id=share_id)

    def pop(self, remotepath: str = "", topdown: Iterable[str] = []) -> Optional[PcsFile]:
        """Pop a node which is at the path of `remotepath` or `topdown`"""

        if not topdown:
            assert remotepath.startswith("/")
            topdown = split_posix_path(remotepath)

        parts = list(topdown)
        dest = parts[-1]
        parent = parts[:-1]
        assert len(parent) > 0, "NO pop root"

        node = self._dfs(list(parent), self.root, pull=False)
        if node:
            if dest in node.sub_nodes:
                sub_node = node.pop(dest)
                if sub_node:
                    self._pop(sub_node.file_id)
                    return deepcopy(sub_node.pcs_file)
        return None

    def pop_by_file_id(self, file_id: str) -> Optional[PcsFile]:
        node = self._file_id_to_node.get(file_id)
        if not node:
            return None

        return self.pop(node.pcs_file.path)



from alipcs_py.config import AppConfig
from alipcs_py.storage.tables import (
    connect_sqlite,
    bind_tables,
    create_tables,
    modify_table,
    PcsSharedLinkInfoTable,
    PcsFileTable,
)
from alipcs_py.commands.env import CONFIG_PATH, SHARED_STORE_PATH


def init_config(app_config: AppConfig):
    # Connect to shared store database
    if app_config.share.store:
        db, migrator = connect_sqlite(str(SHARED_STORE_PATH))
        bind_tables([PcsFileTable, PcsSharedLinkInfoTable], db)
        create_tables([PcsFileTable, PcsSharedLinkInfoTable], db)

        modify_table(PcsFileTable, db, migrator)
        modify_table(PcsSharedLinkInfoTable, db, migrator)

    if not CONFIG_PATH.exists():
        save_config(app_config)


def save_config(app_config: AppConfig):
    app_config.dump(str(CONFIG_PATH))



from typing import Optional
from pathlib import Path
from os import PathLike
import logging
from logging import Logger

from typing_extensions import Literal, Final


TLogLevel = Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
LogLevels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
DEFAULT_LOG_LEVEL: Final = "ERROR"

_LOG_FORMAT = "%(asctime)-15s | %(levelname)s | %(module)s: %(message)s"


def get_logger(
    name: str,
    fmt: str = _LOG_FORMAT,
    filename: Optional[PathLike] = None,
    level: TLogLevel = DEFAULT_LOG_LEVEL,
) -> Logger:
    logger = logging.getLogger(name)
    logger.setLevel(level)

    stream_handler = logging.StreamHandler()  # stdout
    stream_handler.setFormatter(logging.Formatter(fmt))
    logger.addHandler(stream_handler)

    if filename:
        filename = Path(filename)
        _dir = filename.parent
        if not _dir.exists():
            _dir.mkdir()

        file_handler = logging.FileHandler(filename)
        file_handler.setFormatter(logging.Formatter(fmt))
        logger.addHandler(file_handler)

    return logger



import socket


def avail_port(port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("localhost", port)) != 0


def random_avail_port() -> int:
    """Find a random available port using port 0

    https://www.lifewire.com/port-0-in-tcp-and-udp-818145
    Port 0 is a wildcard port that tells the system to find a suitable port number.
    Unix, Windows, and other operating systems vary in the handling of port 0, but
    the same general convention applies.
    """

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("127.0.0.1", 0))
        s.listen()
        _, port = s.getsockname()
        return port



from typing import Optional

from rich.progress import (
    Progress,
    SpinnerColumn,
    TextColumn,
    BarColumn,
    DownloadColumn,
    TransferSpeedColumn,
    TimeRemainingColumn,
    TaskID,
)
from rich.table import Column

_progress = Progress(
    SpinnerColumn(),
    TextColumn("[bold blue]{task.fields[title]}", justify="right", table_column=Column(overflow="fold")),
    BarColumn(bar_width=40),
    "[progress.percentage]{task.percentage:>3.1f}%",
    "•",
    DownloadColumn(binary_units=True),
    "•",
    TransferSpeedColumn(),
    "•",
    TimeRemainingColumn(),
)


def init_progress_bar():
    if not _progress.live._started:
        _progress.start()


def exit_progress_bar():
    if _progress.live._started:
        _progress.stop()


def progress_task_exists(task_id: Optional[TaskID]) -> bool:
    if task_id is None:
        return False
    return task_id in _progress.task_ids


def remove_progress_task(task_id: Optional[TaskID]):
    if task_id is not None and progress_task_exists(task_id):
        _progress.remove_task(task_id)


def reset_progress_task(task_id: Optional[TaskID]):
    if task_id is not None and progress_task_exists(task_id):
        _progress.reset(task_id)



from typing import Optional, List, Any, Callable
from os import PathLike
from pathlib import Path
from threading import Semaphore
from concurrent.futures import ThreadPoolExecutor, as_completed, Future

from alipcs_py.common.constant import CPU_NUM
from alipcs_py.common.io import RangeRequestIO
from alipcs_py.common.concurrent import sure_release, retry

from rich.progress import TaskID

DEFAULT_MAX_WORKERS = 5


class MeDownloader(RangeRequestIO):
    _executor: ThreadPoolExecutor
    _semaphore: Semaphore
    _futures: List[Future]

    @classmethod
    def _set_executor(
        cls,
        max_workers: int = CPU_NUM,
    ):
        cls._executor = ThreadPoolExecutor(max_workers=max_workers)
        cls._semaphore = Semaphore(max_workers)
        cls._futures = []

    @classmethod
    def _exit_executor(cls):
        if getattr(cls, "_executor", None):
            as_completed(cls._futures)
            cls._futures = []
            cls._executor.__exit__(None, None, None)

    def __init__(self, *args, max_workers: int = CPU_NUM, **kwargs):
        super().__init__(*args, **kwargs)
        if not getattr(self.__class__, "_executor", None):
            self.__class__._set_executor(max_workers)

    def download(
        self,
        localpath: PathLike,
        task_id: Optional[TaskID],
        continue_: bool = False,
        done_callback: Optional[Callable[[Future], Any]] = None,
        except_callback: Optional[Callable[..., Any]] = None,
    ):
        """
        Download the url content to `localpath`

        The downloading work executing in the class ThreadPoolExecutor

        Args:
            continue_ (bool): If set to True, only downloading the remain content depended on
            the size of `localpath`
        """

        self._localpath = localpath
        self._task_id = task_id
        self._except_callback = except_callback
        self.continue_ = continue_

        cls = self.__class__
        cls._semaphore.acquire()

        fut = cls._executor.submit(
            sure_release,
            cls._semaphore,
            self.work,
        )
        if done_callback:
            fut.add_done_callback(done_callback)
        cls._futures.append(fut)

    def _init_fd(self):
        if self.continue_:
            _path = Path(self._localpath)
            if self.seekable():
                _offset = _path.stat().st_size if _path.exists() else 0
                _fd = _path.open("ab")
                _fd.seek(_offset, 0)
            else:
                _offset = 0
                _fd = _path.open("wb")
        else:
            _offset = 0
            _fd = open(self._localpath, "wb")

        self._offset = _offset
        self._fd = _fd

    @retry(30)
    def work(self):
        self._init_fd()

        try:
            start, end = self._offset, len(self)

            for b in self._auto_decrypt_request.read((start, end)):
                self._fd.write(b)
                self._offset += len(b)
                # Call callback
                if self._callback:
                    self._callback(self._task_id, self._offset)
        except Exception as err:
            if self._except_callback:
                self._except_callback(self._task_id)
            self.reset()
            raise err
        finally:
            self._fd.close()



from typing import Union, List, Tuple, IO, Callable, Any, cast
import re
import os
import subprocess
import random
from io import BufferedReader
from abc import ABC, abstractmethod
from zlib import crc32
import hashlib
from hashlib import md5, sha1
import base64

from passlib.crypto.digest import pbkdf1

import ecdsa
from ecdsa import SigningKey, VerifyingKey

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.padding import PKCS7
from cryptography.hazmat.backends import default_backend

from alipcs_py.common import constant
from alipcs_py.common.platform import IS_LINUX, IS_MACOS
from alipcs_py.common.simple_cipher import SimpleCryptography as _SimpleCryptography


def _md5_cmd(localpath: str) -> List[str]:
    if IS_MACOS:
        cmd = ["md5", localpath]
    elif IS_LINUX:
        cmd = ["md5sum", localpath]
    else:  # windows
        cmd = ["CertUtil", "-hashfile", localpath, "MD5"]
    return cmd


def calc_file_md5(localpath: str) -> str:
    cp = subprocess.run(_md5_cmd(localpath), universal_newlines=True, stdout=subprocess.PIPE)

    output = cp.stdout.strip()
    if IS_MACOS:
        return re.split(r"\s+", output)[-1]
    elif IS_LINUX:
        return re.split(r"\s+", output)[0]
    else:  # windows
        cn = output.split("CertUtil")[0].strip()
        cn = cn.split(":")[-1].strip().replace(" ", "")
        return cn


ReadableBuffer = Union[bytes, bytearray]  # stable


def calc_crc32_and_md5(stream: IO, chunk_size: int) -> Tuple[int, str]:
    md5_v = md5()
    crc32_v = 0
    while True:
        buf = stream.read(chunk_size)
        if buf:
            md5_v.update(buf)
            crc32_v = crc32(buf, crc32_v).conjugate()
        else:
            break
    return crc32_v.conjugate() & 0xFFFFFFFF, md5_v.hexdigest()


def calc_hash(hash_method: Callable, buf: Union[str, bytes, IO, BufferedReader], encoding="utf-8") -> str:
    assert isinstance(buf, (str, bytes, IO, BufferedReader))

    if isinstance(buf, str):
        buf = buf.encode(encoding)
        return hash_method(buf).hexdigest()
    elif isinstance(buf, bytes):
        return hash_method(buf).hexdigest()
    else:
        hasher = hash_method()
        while True:
            chunk = buf.read(constant.OneM)
            if not chunk:
                return hasher.hexdigest()
            hasher.update(chunk)


def calc_md5(buf: Union[str, bytes, IO, BufferedReader], encoding="utf-8") -> str:
    return calc_hash(md5, buf, encoding=encoding)


def calc_sha1(buf: Union[str, bytes, IO, BufferedReader], encoding="utf-8") -> str:
    return calc_hash(sha1, buf, encoding=encoding)


def calc_proof_code(io: IO, io_len: int, key: str) -> str:
    if io_len == 0:
        return ""

    key_md5 = calc_md5(key)
    offset = int("0x" + key_md5[:16], 16) % io_len
    pre_offset = io.tell()
    io.seek(offset, 0)
    buf = io.read(8) or b""
    io.seek(pre_offset, 0)
    return base64.b64encode(buf).decode("utf-8")


U8_LIST = list(range(1 << 8))


def random_bytes(size: int, seed: Any = None) -> bytes:
    """Generate random bytes"""

    rg = random.Random(seed)
    return bytes(rg.sample(U8_LIST, size))


def random_sys_bytes(size: int) -> bytes:
    """Generate random bytes with `os.urandom`"""

    return os.urandom(size)


def padding_key(key: Union[str, bytes], length: int = 0, value: bytes = b"\xff") -> bytes:
    """padding key with `value`"""

    assert len(value) < 2

    if isinstance(key, str):
        key = key.encode("utf-8")

    assert len(key) <= length

    pad_len = length - len(key)
    if value:
        pad_bytes = value * (pad_len)
    else:
        pad_bytes = random_sys_bytes(pad_len)
    return key + pad_bytes


def padding_size(length: int, block_size: int, ceil: bool = True) -> int:
    """Return minimum the multiple which is large or equal than the `length`

    Args:
        block_size (int): the length of bytes, no the length of bit
    """

    remainder = length % block_size
    if ceil:
        return (block_size - remainder) * int(remainder != 0) + length
    else:
        return length - remainder


def pkcs7_padding(data: bytes, block_size):
    """
    Args:
        block_size (int): the length of bytes, no the length of bit
    """

    padder = PKCS7(block_size * 8).padder()
    return padder.update(data) + padder.finalize()


def pkcs7_unpadding(data: bytes, block_size):
    """
    Args:
        block_size (int): the length of bytes, no the length of bit
    """

    unpadder = PKCS7(block_size * 8).unpadder()
    return unpadder.update(data) + unpadder.finalize()


def generate_salt(size: int = 8) -> bytes:
    return random_sys_bytes(size)


# Generate key and iv with password and salt
# https://security.stackexchange.com/a/117654
# {{{
def generate_key_iv(
    password: bytes, salt: bytes, key_size: int, iv_size: int, algo: str = "md5"
) -> Tuple[bytes, bytes]:
    def hasher(algo: str, data: bytes) -> bytes:
        hashes = {
            "md5": hashlib.md5,
            "sha256": hashlib.sha256,
            "sha512": hashlib.sha512,
        }
        h = hashes[algo]()
        h.update(data)
        return h.digest()

    if algo == "md5":
        temp = pbkdf1("md5", password, salt, 1, 16)
    else:
        temp = b""

    fd = temp
    while len(fd) < key_size + iv_size:
        temp = hasher(algo, temp + password + salt)
        fd += temp

    key = fd[0:key_size]
    iv = fd[key_size : key_size + iv_size]

    return key, iv


# }}}


class Cryptography(ABC):
    @abstractmethod
    def encrypt(self, data: bytes) -> bytes:
        pass

    @abstractmethod
    def decrypt(self, data: bytes) -> bytes:
        pass

    @abstractmethod
    def reset(self):
        pass

    @abstractmethod
    def finalize(self):
        """Finalize encryptor and decryptor, no return data"""


class SimpleCryptography(Cryptography):
    """Simple Cryptography

    This crypto algorithm uses a random uint8 map to transfer an uint8 to another uint8.
    So, the decryption process does not depend on previous decrypted data.

    The algorithm is vulnerable, so NO using to encrypt important data.
    """

    def __init__(self, key):
        self._c = _SimpleCryptography(key)
        self._key = key

    def encrypt(self, data: bytes) -> bytes:
        return self._c.encrypt(data)

    def decrypt(self, data: bytes) -> bytes:
        return self._c.decrypt(data)

    def reset(self):
        pass

    def finalize(self):
        pass


class ChaCha20Cryptography(Cryptography):
    """ChaCha20 Cryptography

    ChaCha20 stream algorithm.

    The decryption process does depend on previous decrypted data.
    """

    def __init__(self, key: bytes, nonce: bytes):
        assert len(key) == 32
        assert len(nonce) == 16

        self._key = key
        self._nonce = nonce
        self.reset()

    def encrypt(self, data: bytes) -> bytes:
        return self._encryptor.update(data)

    def decrypt(self, data: bytes) -> bytes:
        return self._decryptor.update(data)

    def reset(self):
        cipher = Cipher(
            algorithms.ChaCha20(self._key, self._nonce),
            mode=None,
            backend=default_backend(),
        )
        self._encryptor = cipher.encryptor()
        self._decryptor = cipher.decryptor()

    def finalize(self):
        self._encryptor.finalize()
        self._decryptor.finalize()


class AES256CBCCryptography(Cryptography):
    def __init__(self, key: bytes, iv: bytes):
        assert len(key) == 32
        assert len(iv) == 16

        self._key = key
        self._iv = iv
        self._mode = modes.CBC(iv)
        self.reset()

    def encrypt(self, data: bytes) -> bytes:
        assert len(data) % 16 == 0
        return self._encryptor.update(data)

    def decrypt(self, data: bytes) -> bytes:
        return self._decryptor.update(data)

    def reset(self):
        cipher = Cipher(algorithms.AES(self._key), mode=self._mode)
        self._encryptor = cipher.encryptor()
        self._decryptor = cipher.decryptor()

    def finalize(self):
        self._encryptor.finalize()
        self._decryptor.finalize()


def aes256cbc_encrypt(data: bytes, key: bytes, iv: bytes):
    crypto = AES256CBCCryptography(key, iv)
    return crypto.encrypt(data) + crypto._encryptor.finalize()


def aes256cbc_decrypt(data: bytes, key: bytes, iv: bytes):
    crypto = AES256CBCCryptography(key, iv)
    return crypto.decrypt(data) + crypto._decryptor.finalize()


def generate_secp256k1_keys() -> Tuple[SigningKey, VerifyingKey]:
    private_key = ecdsa.SigningKey.generate(curve=ecdsa.SECP256k1)
    public_key: VerifyingKey = cast(VerifyingKey, private_key.verifying_key)
    return private_key, public_key
